#    (c) Copyright 2016 IBM Corporation
#    (c) Copyright 2023 SysEleven GmbH
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from oslo_config import cfg
from oslo_utils import importutils

from neutron_vpnaas.api.rpc.agentnotifiers import vpn_rpc_agent_api as nfy_api
from neutron_vpnaas.db.vpn import vpn_agentschedulers_db as agent_db
from neutron_vpnaas.db.vpn.vpn_db import VPNPluginDb
from neutron_vpnaas.db.vpn import vpn_ext_gw_db
from neutron_vpnaas.services.vpn.common import constants
from neutron_vpnaas.services.vpn.ovn import agent_monitor
from neutron_vpnaas.services.vpn.plugin import VPNDriverPlugin


class VPNOVNPlugin(VPNPluginDb,
                   vpn_ext_gw_db.VPNExtGWPlugin_db,
                   agent_db.AZVPNAgentSchedulerDbMixin,
                   agent_monitor.OVNVPNAgentMonitor):
    """Implementation of the VPN Service Plugin.

    This class manages the workflow of VPNaaS request/response.
    Most DB related works are implemented in class
    vpn_db.VPNPluginDb.
    """
    def __init__(self):
        self.vpn_scheduler = importutils.import_object(
            cfg.CONF.vpn_scheduler_driver)
        self.add_periodic_vpn_agent_status_check()
        self.agent_notifiers[constants.AGENT_TYPE_VPN] = \
            nfy_api.VPNAgentNotifyAPI()
        super().__init__()
        registry.subscribe(self.post_fork_initialize,
                           resources.PROCESS,
                           events.AFTER_INIT)

    def check_router_in_use(self, context, router_id):
        pass

    def post_fork_initialize(self, resource, event, trigger, payload=None):
        self.watch_agent_events()

    def vpn_router_agent_binding_changed(self, context, router_id, host):
        pass

    supported_extension_aliases = ["vpnaas",
                                   "vpn-endpoint-groups",
                                   "service-type",
                                   "vpn-agent-scheduler",
                                   "vpn-aes-ctr"]
    path_prefix = "/vpn"


class VPNOVNDriverPlugin(VPNOVNPlugin, VPNDriverPlugin):
    def vpn_router_agent_binding_changed(self, context, router_id, host):
        super().vpn_router_agent_binding_changed(context, router_id, host)
        filters = {'router_id': [router_id]}
        vpnservices = self.get_vpnservices(context, filters=filters)
        for vpnservice in vpnservices:
            driver = self._get_driver_for_vpnservice(context, vpnservice)
            driver.update_port_bindings(context, router_id, host)



#    (c) Copyright 2016 IBM Corporation, All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import random

from neutron.extensions import availability_zone as az_ext
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_log import log as logging

from neutron_vpnaas.extensions import vpn_agentschedulers

LOG = logging.getLogger(__name__)


class VPNScheduler(metaclass=abc.ABCMeta):
    @property
    def l3_plugin(self):
        return directory.get_plugin(plugin_constants.L3)

    @abc.abstractmethod
    def schedule(self, plugin, context, router_id,
                 candidates=None, hints=None):
        """Schedule the router to an active VPN agent.

        Schedule the router only if it is not already scheduled.
        """
        pass

    def _get_unscheduled_routers(self, context, plugin, router_ids=None):
        """Get the list of routers with VPN services to be scheduled.

        If router IDs are omitted, look for all unscheduled routers.

        :param context: the context
        :param plugin: the core plugin
        :param router_ids: the list of routers to be checked for scheduling
        :returns: the list of routers to be scheduled
        """
        unscheduled_router_ids = plugin.get_unscheduled_vpn_routers(
            context, router_ids=router_ids)
        if unscheduled_router_ids:
            return self.l3_plugin.get_routers(
                context, filters={'id': unscheduled_router_ids})
        return []

    def _get_routers_can_schedule(self, context, plugin, routers, vpn_agent):
        """Get the subset of routers whose VPN services can be scheduled on
        the VPN agent.
        """
        # Assuming that only an active, enabled VPN agent is passed in,
        # all routers can be scheduled to it
        return routers

    def auto_schedule_routers(self, plugin, context, vpn_agent):
        """Schedule non-hosted routers to a VPN agent.

        :returns: True if routers have been successfully assigned to the agent
        """
        unscheduled_routers = self._get_unscheduled_routers(context, plugin)

        target_routers = self._get_routers_can_schedule(
            context, plugin, unscheduled_routers, vpn_agent)
        if not target_routers:
            if unscheduled_routers:
                LOG.warning('No unscheduled routers compatible with VPN agent '
                            'configuration on host %s', vpn_agent['host'])
            return []

        self._bind_routers(context, plugin, target_routers, vpn_agent)
        return [router['id'] for router in target_routers]

    def _get_candidates(self, plugin, context, sync_router):
        """Return VPN agents where a router could be scheduled."""
        active_vpn_agents = plugin.get_vpn_agents(context, active=True)
        if not active_vpn_agents:
            LOG.warning('No active VPN agents')
        return active_vpn_agents

    def _bind_routers(self, context, plugin, routers, vpn_agent):
        for router in routers:
            plugin.create_router_to_agent_binding(
                context, router['id'], vpn_agent['id'])

    def _schedule_router(self, plugin, context, router_id,
                         candidates=None):
        current_vpn_agents = plugin.get_vpn_agents_hosting_routers(
            context, [router_id])
        if current_vpn_agents:
            chosen_agent = current_vpn_agents[0]
            LOG.debug('VPN service of router %(router_id)s has already '
                      'been hosted by VPN agent %(agent_id)s',
                      {'router_id': router_id,
                       'agent_id': chosen_agent})
            return chosen_agent

        sync_router = self.l3_plugin.get_router(context, router_id)
        candidates = candidates or self._get_candidates(
            plugin, context, sync_router)
        if not candidates:
            raise vpn_agentschedulers.RouterReschedulingFailed(
                router_id=router_id)

        chosen_agent = self._choose_vpn_agent(plugin, context, candidates)
        if plugin.create_router_to_agent_binding(context, router_id,
                                                 chosen_agent['id']):
            return chosen_agent

    @abc.abstractmethod
    def _choose_vpn_agent(self, plugin, context, candidates):
        """Choose an agent from candidates based on a specific policy."""
        pass


class ChanceScheduler(VPNScheduler):
    """Randomly allocate an VPN agent for a router."""

    def schedule(self, plugin, context, router_id,
                 candidates=None):
        return self._schedule_router(
            plugin, context, router_id, candidates=candidates)

    def _choose_vpn_agent(self, plugin, context, candidates):
        return random.choice(candidates)


class LeastRoutersScheduler(VPNScheduler):
    """Allocate to an VPN agent with the least number of routers bound."""

    def schedule(self, plugin, context, router_id,
                 candidates=None):
        return self._schedule_router(
            plugin, context, router_id, candidates=candidates)

    def _choose_vpn_agent(self, plugin, context, candidates):
        candidates_dict = {c['id']: c for c in candidates}
        chosen_agent_id = plugin.get_vpn_agent_with_min_routers(
            context, candidates_dict.keys())
        return candidates_dict[chosen_agent_id]


class AZLeastRoutersScheduler(LeastRoutersScheduler):
    """Availability zone aware scheduler."""
    def _get_az_hints(self, router):
        return (router.get(az_ext.AZ_HINTS) or
                cfg.CONF.default_availability_zones)

    def _get_routers_can_schedule(self, context, plugin, routers, vpn_agent):
        """Overwrite VPNScheduler's method to filter by availability zone."""
        target_routers = []
        for r in routers:
            az_hints = self._get_az_hints(r)
            if not az_hints or vpn_agent['availability_zone'] in az_hints:
                target_routers.append(r)

        if not target_routers:
            return

        return super()._get_routers_can_schedule(
            context, plugin, target_routers, vpn_agent)

    def _get_candidates(self, plugin, context, sync_router):
        """Overwrite VPNScheduler's method to filter by availability zone."""
        all_candidates = super()._get_candidates(plugin, context, sync_router)

        candidates = []
        az_hints = self._get_az_hints(sync_router)
        for agent in all_candidates:
            if not az_hints or agent['availability_zone'] in az_hints:
                candidates.append(agent)

        return candidates



#    (c) Copyright 2013 Hewlett-Packard Development Company, L.P.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from neutron.tests.unit.api.v2 import test_base as test_api_v2
from neutron_lib.plugins import constants as nconstants
from oslo_utils import uuidutils
from webob import exc

from neutron_vpnaas.extensions import vpnaas
from neutron_vpnaas.tests import base

_uuid = uuidutils.generate_uuid
_get_path = test_api_v2._get_path


class VpnaasExtensionTestCase(base.ExtensionTestCase):
    fmt = 'json'

    def setUp(self):
        super().setUp()
        plural_mappings = {'ipsecpolicy': 'ipsecpolicies',
                           'ikepolicy': 'ikepolicies',
                           'ipsec_site_connection': 'ipsec-site-connections',
                           'endpoint_group': 'endpoint-groups'}
        self.setup_extension(
            'neutron_vpnaas.extensions.vpnaas.VPNPluginBase', nconstants.VPN,
            vpnaas.Vpnaas, 'vpn', plural_mappings=plural_mappings,
            use_quota=True)

    def test_ikepolicy_create(self):
        """Test case to create an ikepolicy."""
        ikepolicy_id = _uuid()
        data = {'ikepolicy': {'name': 'ikepolicy1',
                              'description': 'myikepolicy1',
                              'auth_algorithm': 'sha1',
                              'encryption_algorithm': 'aes-128',
                              'phase1_negotiation_mode': 'main',
                              'lifetime': {
                                  'units': 'seconds',
                                  'value': 3600},
                              'ike_version': 'v1',
                              'pfs': 'group5',
                              'tenant_id': _uuid()}}

        return_value = copy.copy(data['ikepolicy'])
        return_value.update({'id': ikepolicy_id})

        instance = self.plugin.return_value
        instance.create_ikepolicy.return_value = return_value
        res = self.api.post(_get_path('vpn/ikepolicies', fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt)
        self.assertEqual(1, instance.create_ikepolicy.call_count)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('ikepolicy', res)
        self.assertDictSupersetOf(return_value, res['ikepolicy'])

    def test_ikepolicy_create_with_aggressive_mode(self):
        """Test case to create an ikepolicy with agressive mode."""
        ikepolicy_id = _uuid()
        data = {'ikepolicy': {'name': 'ikepolicy1',
                              'description': 'myikepolicy1',
                              'auth_algorithm': 'sha1',
                              'encryption_algorithm': 'aes-128',
                              'phase1_negotiation_mode': 'aggressive',
                              'lifetime': {
                                  'units': 'seconds',
                                  'value': 3600},
                              'ike_version': 'v1',
                              'pfs': 'group5',
                              'tenant_id': _uuid()}}

        return_value = copy.copy(data['ikepolicy'])
        return_value.update({'id': ikepolicy_id})

        instance = self.plugin.return_value
        instance.create_ikepolicy.return_value = return_value
        res = self.api.post(_get_path('vpn/ikepolicies', fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt)
        self.assertEqual(1, instance.create_ikepolicy.call_count)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('ikepolicy', res)
        self.assertDictSupersetOf(return_value, res['ikepolicy'])

    def test_ikepolicy_list(self):
        """Test case to list all ikepolicies."""
        ikepolicy_id = _uuid()
        return_value = [{'name': 'ikepolicy1',
                         'auth_algorithm': 'sha1',
                         'encryption_algorithm': 'aes-128',
                         'pfs': 'group5',
                         'ike_version': 'v1',
                         'id': ikepolicy_id}]

        instance = self.plugin.return_value
        instance.get_ikepolicies.return_value = return_value

        res = self.api.get(_get_path('vpn/ikepolicies', fmt=self.fmt))

        instance.get_ikepolicies.assert_called_with(mock.ANY,
                                                    fields=mock.ANY,
                                                    filters=mock.ANY)
        self.assertEqual(exc.HTTPOk.code, res.status_int)

    def test_ikepolicy_update(self):
        """Test case to update an ikepolicy."""
        ikepolicy_id = _uuid()
        update_data = {'ikepolicy': {'name': 'ikepolicy1',
                                     'encryption_algorithm': 'aes-256'}}
        return_value = {'name': 'ikepolicy1',
                        'auth_algorithm': 'sha1',
                        'encryption_algorithm': 'aes-256',
                        'phase1_negotiation_mode': 'main',
                        'lifetime': {
                            'units': 'seconds',
                            'value': 3600},
                        'ike_version': 'v1',
                        'pfs': 'group5',
                        'tenant_id': _uuid(),
                        'id': ikepolicy_id}

        instance = self.plugin.return_value
        instance.update_ikepolicy.return_value = return_value

        res = self.api.put(_get_path('vpn/ikepolicies', id=ikepolicy_id,
                                     fmt=self.fmt),
                           self.serialize(update_data))

        instance.update_ikepolicy.assert_called_with(mock.ANY, ikepolicy_id,
                                                     ikepolicy=update_data)
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('ikepolicy', res)
        self.assertEqual(return_value, res['ikepolicy'])

    def test_ikepolicy_update_with_aggressive_mode(self):
        """Test case to update an ikepolicy with aggressive mode."""
        ikepolicy_id = _uuid()
        update_data = {'ikepolicy':
                       {'name': 'ikepolicy1',
                        'phase1_negotiation_mode': 'aggressive',
                        'encryption_algorithm': 'aes-256'}}
        return_value = {'name': 'ikepolicy1',
                        'auth_algorithm': 'sha1',
                        'encryption_algorithm': 'aes-256',
                        'phase1_negotiation_mode': 'aggressive',
                        'lifetime': {
                            'units': 'seconds',
                            'value': 3600},
                        'ike_version': 'v1',
                        'pfs': 'group5',
                        'tenant_id': _uuid(),
                        'id': ikepolicy_id}

        instance = self.plugin.return_value
        instance.update_ikepolicy.return_value = return_value

        res = self.api.put(_get_path('vpn/ikepolicies', id=ikepolicy_id,
                                     fmt=self.fmt),
                           self.serialize(update_data))

        instance.update_ikepolicy.assert_called_with(mock.ANY, ikepolicy_id,
                                                     ikepolicy=update_data)
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('ikepolicy', res)
        self.assertEqual(return_value, res['ikepolicy'])

    def test_ikepolicy_get(self):
        """Test case to get or show an ikepolicy."""
        ikepolicy_id = _uuid()
        return_value = {'name': 'ikepolicy1',
                        'auth_algorithm': 'sha1',
                        'encryption_algorithm': 'aes-128',
                        'phase1_negotiation_mode': 'main',
                        'lifetime': {
                            'units': 'seconds',
                            'value': 3600},
                        'ike_version': 'v1',
                        'pfs': 'group5',
                        'tenant_id': _uuid(),
                        'id': ikepolicy_id}

        instance = self.plugin.return_value
        instance.get_ikepolicy.return_value = return_value

        res = self.api.get(_get_path('vpn/ikepolicies', id=ikepolicy_id,
                                     fmt=self.fmt))

        instance.get_ikepolicy.assert_called_with(mock.ANY,
                                                  ikepolicy_id,
                                                  fields=mock.ANY)
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('ikepolicy', res)
        self.assertEqual(return_value, res['ikepolicy'])

    def test_ikepolicy_delete(self):
        """Test case to delete an ikepolicy."""
        self._test_entity_delete('ikepolicy')

    def test_ipsecpolicy_create(self):
        """Test case to create an ipsecpolicy."""
        ipsecpolicy_id = _uuid()
        data = {'ipsecpolicy': {'name': 'ipsecpolicy1',
                                'description': 'myipsecpolicy1',
                                'auth_algorithm': 'sha1',
                                'encryption_algorithm': 'aes-128',
                                'encapsulation_mode': 'tunnel',
                                'lifetime': {
                                    'units': 'seconds',
                                    'value': 3600},
                                'transform_protocol': 'esp',
                                'pfs': 'group5',
                                'tenant_id': _uuid()}}
        return_value = copy.copy(data['ipsecpolicy'])
        return_value.update({'id': ipsecpolicy_id})

        instance = self.plugin.return_value
        instance.create_ipsecpolicy.return_value = return_value
        res = self.api.post(_get_path('vpn/ipsecpolicies', fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt)
        self.assertEqual(1, instance.create_ipsecpolicy.call_count)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('ipsecpolicy', res)
        self.assertDictSupersetOf(return_value, res['ipsecpolicy'])

    def test_ipsecpolicy_list(self):
        """Test case to list an ipsecpolicy."""
        ipsecpolicy_id = _uuid()
        return_value = [{'name': 'ipsecpolicy1',
                         'auth_algorithm': 'sha1',
                         'encryption_algorithm': 'aes-128',
                         'pfs': 'group5',
                         'id': ipsecpolicy_id}]

        instance = self.plugin.return_value
        instance.get_ipsecpolicies.return_value = return_value

        res = self.api.get(_get_path('vpn/ipsecpolicies', fmt=self.fmt))

        instance.get_ipsecpolicies.assert_called_with(mock.ANY,
                                                      fields=mock.ANY,
                                                      filters=mock.ANY)
        self.assertEqual(exc.HTTPOk.code, res.status_int)

    def test_ipsecpolicy_update(self):
        """Test case to update an ipsecpolicy."""
        ipsecpolicy_id = _uuid()
        update_data = {'ipsecpolicy': {'name': 'ipsecpolicy1',
                                       'encryption_algorithm': 'aes-256'}}
        return_value = {'name': 'ipsecpolicy1',
                        'auth_algorithm': 'sha1',
                        'encryption_algorithm': 'aes-128',
                        'encapsulation_mode': 'tunnel',
                        'lifetime': {
                            'units': 'seconds',
                            'value': 3600},
                        'transform_protocol': 'esp',
                        'pfs': 'group5',
                        'tenant_id': _uuid(),
                        'id': ipsecpolicy_id}

        instance = self.plugin.return_value
        instance.update_ipsecpolicy.return_value = return_value

        res = self.api.put(_get_path('vpn/ipsecpolicies',
                                     id=ipsecpolicy_id,
                                     fmt=self.fmt),
                           self.serialize(update_data))

        instance.update_ipsecpolicy.assert_called_with(mock.ANY,
                                                       ipsecpolicy_id,
                                                       ipsecpolicy=update_data)
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('ipsecpolicy', res)
        self.assertEqual(return_value, res['ipsecpolicy'])

    def test_ipsecpolicy_get(self):
        """Test case to get or show an ipsecpolicy."""
        ipsecpolicy_id = _uuid()
        return_value = {'name': 'ipsecpolicy1',
                        'auth_algorithm': 'sha1',
                        'encryption_algorithm': 'aes-128',
                        'encapsulation_mode': 'tunnel',
                        'lifetime': {
                            'units': 'seconds',
                            'value': 3600},
                        'transform_protocol': 'esp',
                        'pfs': 'group5',
                        'tenant_id': _uuid(),
                        'id': ipsecpolicy_id}

        instance = self.plugin.return_value
        instance.get_ipsecpolicy.return_value = return_value

        res = self.api.get(_get_path('vpn/ipsecpolicies',
                                     id=ipsecpolicy_id,
                                     fmt=self.fmt))

        instance.get_ipsecpolicy.assert_called_with(mock.ANY,
                                                    ipsecpolicy_id,
                                                    fields=mock.ANY)
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('ipsecpolicy', res)
        self.assertEqual(return_value, res['ipsecpolicy'])

    def test_ipsecpolicy_delete(self):
        """Test case to delete an ipsecpolicy."""
        self._test_entity_delete('ipsecpolicy')

    def _test_vpnservice_create(self, more_args, defaulted_args):
        """Helper to test VPN service creation.

        Allows additional args to be specified for different test cases.
        Includes expected args, for case where an optional args are not
        specified and API applies defaults.
        """

        data = {'vpnservice': {'name': 'vpnservice1',
                               'description': 'descr_vpn1',
                               'router_id': _uuid(),
                               'admin_state_up': True,
                               'tenant_id': _uuid()}}
        data['vpnservice'].update(more_args)

        # Add in any default values for args that were not provided
        actual_args = copy.copy(data)
        actual_args['vpnservice'].update(defaulted_args)

        return_value = copy.copy(data['vpnservice'])
        return_value.update({'status': "ACTIVE", 'id': _uuid()})
        return_value.update(defaulted_args)
        instance = self.plugin.return_value
        instance.create_vpnservice.return_value = return_value

        res = self.api.post(_get_path('vpn/vpnservices', fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt)
        self.assertEqual(1, instance.create_vpnservice.call_count)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('vpnservice', res)
        self.assertDictSupersetOf(return_value, res['vpnservice'])

    def test_vpnservice_create(self):
        """Create VPN service using subnet (older API)."""
        subnet = {'subnet_id': _uuid()}
        self._test_vpnservice_create(more_args=subnet, defaulted_args={})

    def test_vpnservice_create_no_subnet(self):
        """Test case to create a vpnservice w/o subnet (newer API)."""
        no_subnet = {'subnet_id': None}
        self._test_vpnservice_create(more_args={}, defaulted_args=no_subnet)

    def test_vpnservice_list(self):
        """Test case to list all vpnservices."""
        vpnservice_id = _uuid()
        return_value = [{'name': 'vpnservice1',
                         'tenant_id': _uuid(),
                         'status': 'ACTIVE',
                         'id': vpnservice_id}]

        instance = self.plugin.return_value
        instance.get_vpnservice.return_value = return_value

        res = self.api.get(_get_path('vpn/vpnservices', fmt=self.fmt))

        instance.get_vpnservices.assert_called_with(mock.ANY,
                                                    fields=mock.ANY,
                                                    filters=mock.ANY)
        self.assertEqual(exc.HTTPOk.code, res.status_int)

    def test_vpnservice_update(self):
        """Test case to update a vpnservice."""
        vpnservice_id = _uuid()
        update_data = {'vpnservice': {'admin_state_up': False}}
        return_value = {'name': 'vpnservice1',
                        'admin_state_up': False,
                        'subnet_id': _uuid(),
                        'router_id': _uuid(),
                        'tenant_id': _uuid(),
                        'status': "ACTIVE",
                        'id': vpnservice_id}

        instance = self.plugin.return_value
        instance.update_vpnservice.return_value = return_value

        res = self.api.put(_get_path('vpn/vpnservices',
                                     id=vpnservice_id,
                                     fmt=self.fmt),
                           self.serialize(update_data))

        instance.update_vpnservice.assert_called_with(mock.ANY,
                                                      vpnservice_id,
                                                      vpnservice=update_data)
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('vpnservice', res)
        self.assertEqual(return_value, res['vpnservice'])

    def test_vpnservice_get(self):
        """Test case to get or show a vpnservice."""
        vpnservice_id = _uuid()
        return_value = {'name': 'vpnservice1',
                        'admin_state_up': True,
                        'subnet_id': _uuid(),
                        'router_id': _uuid(),
                        'tenant_id': _uuid(),
                        'status': "ACTIVE",
                        'id': vpnservice_id}

        instance = self.plugin.return_value
        instance.get_vpnservice.return_value = return_value

        res = self.api.get(_get_path('vpn/vpnservices',
                                     id=vpnservice_id,
                                     fmt=self.fmt))

        instance.get_vpnservice.assert_called_with(mock.ANY,
                                                   vpnservice_id,
                                                   fields=mock.ANY)
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('vpnservice', res)
        self.assertEqual(return_value, res['vpnservice'])

    def test_vpnservice_delete(self):
        """Test case to delete a vpnservice."""
        self._test_entity_delete('vpnservice')

    def _test_ipsec_site_connection_create(self, more_args, defaulted_args):
        """Helper to test creating IPSec connection."""
        ipsecsite_con_id = _uuid()
        ikepolicy_id = _uuid()
        ipsecpolicy_id = _uuid()
        data = {
            'ipsec_site_connection': {'name': 'connection1',
                                      'description': 'Remote-connection1',
                                      'peer_address': '192.168.1.10',
                                      'peer_id': '192.168.1.10',
                                      'mtu': 1500,
                                      'psk': 'abcd',
                                      'initiator': 'bi-directional',
                                      'dpd': {
                                          'action': 'hold',
                                          'interval': 30,
                                          'timeout': 120},
                                      'ikepolicy_id': ikepolicy_id,
                                      'ipsecpolicy_id': ipsecpolicy_id,
                                      'vpnservice_id': _uuid(),
                                      'admin_state_up': True,
                                      'tenant_id': _uuid()}
        }
        data['ipsec_site_connection'].update(more_args)

        # Add in any default values for args that were not provided
        actual_args = copy.copy(data)
        actual_args['ipsec_site_connection'].update(defaulted_args)

        return_value = copy.copy(data['ipsec_site_connection'])
        return_value.update({'status': "ACTIVE", 'id': ipsecsite_con_id})
        return_value.update(defaulted_args)
        instance = self.plugin.return_value
        instance.create_ipsec_site_connection.return_value = return_value

        res = self.api.post(_get_path('vpn/ipsec-site-connections',
                                      fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt)
        self.assertEqual(1, instance.create_ipsec_site_connection.call_count)
        self.assertEqual(exc.HTTPCreated.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('ipsec_site_connection', res)
        self.assertDictSupersetOf(return_value, res['ipsec_site_connection'])

    def test_ipsec_site_connection_create(self):
        """Create an IPSec connection with peer CIDRs (old API)."""
        more_args = {'peer_cidrs': ['192.168.2.0/24', '192.168.3.0/24'],
                     'local_id': ''}
        no_endpoint_groups = {'local_ep_group_id': None,
                              'peer_ep_group_id': None}
        self._test_ipsec_site_connection_create(
            more_args=more_args, defaulted_args=no_endpoint_groups)

    def test_ipsec_site_connection_create_with_endpoints(self):
        """Create an IPSec connection with endpoint groups (new API)."""
        more_args = {'local_ep_group_id': _uuid(),
                     'peer_ep_group_id': _uuid(),
                     'local_id': ''}
        no_peer_cidrs = {'peer_cidrs': []}
        self._test_ipsec_site_connection_create(more_args=more_args,
                                                defaulted_args=no_peer_cidrs)

    def test_ipsec_site_connection_create_with_invalid_cidr_format(self):
        peer_cidrs = ['192.168.2.0/24', '10/8']
        data = {
            'ipsec_site_connection': {'name': 'connection1',
                                      'description': 'Remote-connection1',
                                      'peer_address': '192.168.1.10',
                                      'peer_id': '192.168.1.10',
                                      'peer_cidrs': peer_cidrs,
                                      'mtu': 1500,
                                      'psk': 'abcd',
                                      'initiator': 'bi-directional',
                                      'dpd': {
                                          'action': 'hold',
                                          'interval': 30,
                                          'timeout': 120},
                                      'ikepolicy_id': _uuid(),
                                      'ipsecpolicy_id': _uuid(),
                                      'vpnservice_id': _uuid(),
                                      'admin_state_up': True,
                                      'tenant_id': _uuid()}
        }
        res = self.api.post(_get_path('vpn/ipsec-site-connections',
                                      fmt=self.fmt),
                            self.serialize(data),
                            content_type='application/%s' % self.fmt,
                            expect_errors=True)
        self.assertEqual(exc.HTTPBadRequest.code, res.status_int)

    def test_ipsec_site_connection_list(self):
        """Test case to list all ipsec_site_connections."""
        ipsecsite_con_id = _uuid()
        return_value = [{'name': 'connection1',
                         'peer_address': '192.168.1.10',
                         'peer_cidrs': ['192.168.2.0/24', '192.168.3.0/24'],
                         'route_mode': 'static',
                         'auth_mode': 'psk',
                         'local_ep_group_id': None,
                         'peer_ep_group_id': None,
                         'tenant_id': _uuid(),
                         'status': 'ACTIVE',
                         'id': ipsecsite_con_id}]

        instance = self.plugin.return_value
        instance.get_ipsec_site_connections.return_value = return_value

        res = self.api.get(
            _get_path('vpn/ipsec-site-connections', fmt=self.fmt))

        instance.get_ipsec_site_connections.assert_called_with(
            mock.ANY, fields=mock.ANY, filters=mock.ANY
        )
        self.assertEqual(exc.HTTPOk.code, res.status_int)

    def test_ipsec_site_connection_update(self):
        """Test case to update a ipsec_site_connection."""
        ipsecsite_con_id = _uuid()
        update_data = {'ipsec_site_connection': {'admin_state_up': False}}
        return_value = {'name': 'connection1',
                        'description': 'Remote-connection1',
                        'peer_address': '192.168.1.10',
                        'peer_id': '192.168.1.10',
                        'peer_cidrs': ['192.168.2.0/24', '192.168.3.0/24'],
                        'mtu': 1500,
                        'psk': 'abcd',
                        'initiator': 'bi-directional',
                        'dpd': {
                            'action': 'hold',
                            'interval': 30,
                            'timeout': 120},
                        'ikepolicy_id': _uuid(),
                        'ipsecpolicy_id': _uuid(),
                        'vpnservice_id': _uuid(),
                        'admin_state_up': False,
                        'local_ep_group_id': None,
                        'peer_ep_group_id': None,
                        'tenant_id': _uuid(),
                        'status': 'ACTIVE',
                        'id': ipsecsite_con_id}

        instance = self.plugin.return_value
        instance.update_ipsec_site_connection.return_value = return_value

        res = self.api.put(_get_path('vpn/ipsec-site-connections',
                                     id=ipsecsite_con_id,
                                     fmt=self.fmt),
                           self.serialize(update_data))

        instance.update_ipsec_site_connection.assert_called_with(
            mock.ANY, ipsecsite_con_id, ipsec_site_connection=update_data
        )

        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('ipsec_site_connection', res)
        self.assertEqual(return_value, res['ipsec_site_connection'])

    def test_ipsec_site_connection_get(self):
        """Test case to get or show a ipsec_site_connection."""
        ipsecsite_con_id = _uuid()
        return_value = {'name': 'connection1',
                        'description': 'Remote-connection1',
                        'peer_address': '192.168.1.10',
                        'peer_id': '192.168.1.10',
                        'peer_cidrs': ['192.168.2.0/24',
                                       '192.168.3.0/24'],
                        'mtu': 1500,
                        'psk': 'abcd',
                        'initiator': 'bi-directional',
                        'dpd': {
                            'action': 'hold',
                            'interval': 30,
                            'timeout': 120},
                        'ikepolicy_id': _uuid(),
                        'ipsecpolicy_id': _uuid(),
                        'vpnservice_id': _uuid(),
                        'admin_state_up': True,
                        'tenant_id': _uuid(),
                        'local_ep_group_id': None,
                        'peer_ep_group_id': None,
                        'status': 'ACTIVE',
                        'id': ipsecsite_con_id}

        instance = self.plugin.return_value
        instance.get_ipsec_site_connection.return_value = return_value

        res = self.api.get(_get_path('vpn/ipsec-site-connections',
                                     id=ipsecsite_con_id,
                                     fmt=self.fmt))

        instance.get_ipsec_site_connection.assert_called_with(
            mock.ANY, ipsecsite_con_id, fields=mock.ANY
        )
        self.assertEqual(exc.HTTPOk.code, res.status_int)
        res = self.deserialize(res)
        self.assertIn('ipsec_site_connection', res)
        self.assertEqual(return_value, res['ipsec_site_connection'])

    def test_ipsec_site_connection_delete(self):
        """Test case to delete a ipsec_site_connection."""
        self._test_entity_delete('ipsec_site_connection')



# Copyright 2016 <Dongcan Ye/Awcloud>
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add_auth_algorithm_sha384_and_sha512

Revision ID: 38893903cbde
Revises: 52783a36bd67
Create Date: 2016-11-04 18:00:49.219140

"""

from alembic import op
from neutron.db import migration
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '38893903cbde'
down_revision = '52783a36bd67'

# milestone identifier, used by neutron-db-manage
neutron_milestone = [migration.OCATA]

new_auth = sa.Enum('sha1', 'sha256', 'sha384', 'sha512',
                   name='vpn_auth_algorithms')


def upgrade():
    op.alter_column('ikepolicies', 'auth_algorithm', type_=new_auth,
                    nullable=False, do_drop=False)
    op.alter_column('ipsecpolicies', 'auth_algorithm', type_=new_auth,
                    nullable=False, do_rename=False, do_create=False)



#    (c) Copyright 2016 IBM Corporation, All Rights Reserved.
#    (c) Copyright 2023 SysEleven GmbH
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.db.models import l3 as l3_models
from neutron.db import models_v2
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as lib_constants
from neutron_lib.db import api as db_api
from neutron_lib.db import model_base
from neutron_lib.db import model_query
from neutron_lib import exceptions as n_exc
from neutron_lib.plugins import constants as plugin_const
from neutron_lib.plugins import directory
from oslo_log import log as logging
from oslo_utils import uuidutils
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.orm import exc

from neutron_vpnaas._i18n import _


LOG = logging.getLogger(__name__)


class RouterIsNotVPNExternal(n_exc.BadRequest):
    message = _("Router %(router_id)s has no VPN external network gateway set")


class RouterHasVPNExternal(n_exc.BadRequest):
    message = _(
        "Router %(router_id)s already has VPN external network gateway")


class VPNNetworkInUse(n_exc.NetworkInUse):
    message = _("Network %(network_id)s is used by VPN service")


class VPNExtGW(model_base.BASEV2, model_base.HasId, model_base.HasProject):
    __tablename__ = 'vpn_ext_gws'
    router_id = sa.Column(sa.String(36), sa.ForeignKey('routers.id'),
                          nullable=False, unique=True)
    status = sa.Column(sa.String(16), nullable=False)
    gw_port_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('ports.id', ondelete='SET NULL'))
    transit_port_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('ports.id', ondelete='SET NULL'))
    transit_network_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('networks.id', ondelete='SET NULL'))
    transit_subnet_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('subnets.id', ondelete='SET NULL'))

    gw_port = orm.relationship(models_v2.Port, lazy='joined',
                               foreign_keys=[gw_port_id])
    transit_port = orm.relationship(models_v2.Port, lazy='joined',
                                    foreign_keys=[transit_port_id])
    transit_network = orm.relationship(models_v2.Network)
    transit_subnet = orm.relationship(models_v2.Subnet)
    router = orm.relationship(l3_models.Router)


@registry.has_registry_receivers
class VPNExtGWPlugin_db:
    """DB class to support vpn external ports configuration."""

    @property
    def _core_plugin(self):
        return directory.get_plugin()

    @property
    def _vpn_plugin(self):
        return directory.get_plugin(plugin_const.VPN)

    @staticmethod
    @registry.receives(resources.PORT, [events.BEFORE_DELETE])
    def _prevent_vpn_port_delete_callback(resource, event,
                                          trigger, payload=None):
        vpn_plugin = directory.get_plugin(plugin_const.VPN)
        if vpn_plugin:
            vpn_plugin.prevent_vpn_port_deletion(payload.context,
                                                 payload.resource_id)

    @db_api.CONTEXT_READER
    def _id_used(self, context, id_column, resource_id):
        return context.session.query(VPNExtGW).filter(
            sa.and_(
                id_column == resource_id,
                VPNExtGW.status != lib_constants.PENDING_DELETE
            )
        ).count() > 0

    def prevent_vpn_port_deletion(self, context, port_id):
        """Checks to make sure a port is allowed to be deleted.

        Raises an exception if this is not the case.  This should be called by
        any plugin when the API requests the deletion of a port, since some
        ports for L3 are not intended to be deleted directly via a DELETE
        to /ports, but rather via other API calls that perform the proper
        deletion checks.
        """
        try:
            port = self._core_plugin.get_port(context, port_id)
        except n_exc.PortNotFound:
            # non-existent ports don't need to be protected from deletion
            return

        port_id_column = {
            lib_constants.DEVICE_OWNER_VPN_ROUTER_GW: VPNExtGW.gw_port_id,
            lib_constants.DEVICE_OWNER_VPN_TRANSIT_NETWORK:
                VPNExtGW.transit_port_id,
        }.get(port['device_owner'])

        if not port_id_column:
            # This is not a VPN port
            return

        if self._id_used(context, port_id_column, port_id):
            reason = _('has device owner %s') % port['device_owner']
            raise n_exc.ServicePortInUse(port_id=port['id'], reason=reason)

    @staticmethod
    @registry.receives(resources.SUBNET, [events.BEFORE_DELETE])
    def _prevent_vpn_subnet_delete_callback(resource, event,
                                            trigger, payload=None):
        vpn_plugin = directory.get_plugin(plugin_const.VPN)
        if vpn_plugin:
            vpn_plugin.prevent_vpn_subnet_deletion(payload.context,
                                                   payload.resource_id)

    def prevent_vpn_subnet_deletion(self, context, subnet_id):
        if self._id_used(context, VPNExtGW.transit_subnet_id, subnet_id):
            reason = _('Subnet is used by VPN service')
            raise n_exc.SubnetInUse(subnet_id=subnet_id, reason=reason)

    @staticmethod
    @registry.receives(resources.NETWORK, [events.BEFORE_DELETE])
    def _prevent_vpn_network_delete_callback(resource, event,
                                             trigger, payload=None):
        vpn_plugin = directory.get_plugin(plugin_const.VPN)
        if vpn_plugin:
            vpn_plugin.prevent_vpn_network_deletion(payload.context,
                                                    payload.resource_id)

    def prevent_vpn_network_deletion(self, context, network_id):
        if self._id_used(context, VPNExtGW.transit_network_id, network_id):
            raise VPNNetworkInUse(network_id=network_id)

    def _make_vpn_ext_gw_dict(self, gateway_db):
        if not gateway_db:
            return None
        gateway = {
            'id': gateway_db['id'],
            'tenant_id': gateway_db['tenant_id'],
            'router_id': gateway_db['router_id'],
            'status': gateway_db['status'],
        }
        if gateway_db.gw_port:
            gateway['network_id'] = gateway_db.gw_port['network_id']
            gateway['external_fixed_ips'] = [
                {'subnet_id': ip["subnet_id"], 'ip_address': ip["ip_address"]}
                for ip in gateway_db.gw_port['fixed_ips']
            ]
        for key in ('gw_port_id', 'transit_port_id', 'transit_network_id',
                    'transit_subnet_id'):
            value = gateway_db.get(key)
            if value:
                gateway[key] = value
        return gateway

    def _get_vpn_gw_by_router_id(self, context, router_id):
        try:
            gateway_db = context.session.query(VPNExtGW).filter(
                VPNExtGW.router_id == router_id).one()
        except exc.NoResultFound:
            return None
        return gateway_db

    @db_api.CONTEXT_READER
    def get_vpn_gw_by_router_id(self, context, router_id):
        return self._get_vpn_gw_by_router_id(context, router_id)

    @db_api.CONTEXT_READER
    def get_vpn_gw_dict_by_router_id(self, context, router_id, refresh=False):
        gateway_db = self._get_vpn_gw_by_router_id(context, router_id)
        if gateway_db and refresh:
            context.session.refresh(gateway_db)
        return self._make_vpn_ext_gw_dict(gateway_db)

    def create_gateway(self, context, gateway):
        info = gateway['gateway']

        with db_api.CONTEXT_WRITER.using(context):
            gateway_db = VPNExtGW(
                id=uuidutils.generate_uuid(),
                tenant_id=info['tenant_id'],
                router_id=info['router_id'],
                status=lib_constants.PENDING_CREATE,
                gw_port_id=info.get('gw_port_id'),
                transit_port_id=info.get('transit_port_id'),
                transit_network_id=info.get('transit_network_id'),
                transit_subnet_id=info.get('transit_subnet_id'))
            context.session.add(gateway_db)

        return self._make_vpn_ext_gw_dict(gateway_db)

    def update_gateway(self, context, gateway_id, gateway):
        info = gateway['gateway']
        with db_api.CONTEXT_WRITER.using(context):
            gateway_db = model_query.get_by_id(context, VPNExtGW, gateway_id)
            gateway_db.update(info)
            return self._make_vpn_ext_gw_dict(gateway_db)

    def delete_gateway(self, context, gateway_id):
        with db_api.CONTEXT_WRITER.using(context):
            query = context.session.query(VPNExtGW)
            return query.filter(VPNExtGW.id == gateway_id).delete()



#    (c) Copyright 2015 Cisco Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Multiple local subnets

Revision ID: 2cb4ee992b41
Revises: 2c82e782d734
Create Date: 2015-09-09 20:32:54.254267

"""

from alembic import op
from oslo_utils import uuidutils
import sqlalchemy as sa
from sqlalchemy.sql import expression as sa_expr

from neutron.db import migration

from neutron_vpnaas.services.vpn.common import constants as v_constants


# revision identifiers, used by Alembic.
revision = '2cb4ee992b41'
down_revision = '2c82e782d734'
depends_on = ('28ee739a7e4b',)

# milestone identifier, used by neutron-db-manage
neutron_milestone = [migration.MITAKA]


vpnservices = sa.Table(
    'vpnservices', sa.MetaData(),
    sa.Column('id', sa.String(length=36), nullable=False),
    sa.Column('tenant_id', sa.String(length=255), nullable=False),
    sa.Column('name', sa.String(255)),
    sa.Column('description', sa.String(255)),
    sa.Column('status', sa.String(16), nullable=False),
    sa.Column('admin_state_up', sa.Boolean(), nullable=False),
    sa.Column('external_v4_ip', sa.String(16)),
    sa.Column('external_v6_ip', sa.String(64)),
    sa.Column('subnet_id', sa.String(36)),
    sa.Column('router_id', sa.String(36), nullable=False))

ipsec_site_conns = sa.Table(
    'ipsec_site_connections', sa.MetaData(),
    sa.Column('id', sa.String(length=36), nullable=False),
    sa.Column('tenant_id', sa.String(length=255), nullable=False),
    sa.Column('name', sa.String(255)),
    sa.Column('description', sa.String(255)),
    sa.Column('peer_address', sa.String(255), nullable=False),
    sa.Column('peer_id', sa.String(255), nullable=False),
    sa.Column('route_mode', sa.String(8), nullable=False),
    sa.Column('mtu', sa.Integer, nullable=False),
    sa.Column('initiator', sa.Enum("bi-directional", "response-only",
                                   name="vpn_initiators"), nullable=False),
    sa.Column('auth_mode', sa.String(16), nullable=False),
    sa.Column('psk', sa.String(255), nullable=False),
    sa.Column('dpd_action', sa.Enum("hold", "clear", "restart", "disabled",
                                    "restart-by-peer", name="vpn_dpd_actions"),
              nullable=False),
    sa.Column('dpd_interval', sa.Integer, nullable=False),
    sa.Column('dpd_timeout', sa.Integer, nullable=False),
    sa.Column('status', sa.String(16), nullable=False),
    sa.Column('admin_state_up', sa.Boolean(), nullable=False),
    sa.Column('vpnservice_id', sa.String(36), nullable=False),
    sa.Column('ipsecpolicy_id', sa.String(36), nullable=False),
    sa.Column('ikepolicy_id', sa.String(36), nullable=False),
    sa.Column('local_ep_group_id', sa.String(36)),
    sa.Column('peer_ep_group_id', sa.String(36)))

ipsecpeercidrs = sa.Table(
    'ipsecpeercidrs', sa.MetaData(),
    sa.Column('cidr', sa.String(32), nullable=False, primary_key=True),
    sa.Column('ipsec_site_connection_id', sa.String(36), primary_key=True))


def _make_endpoint_groups(new_groups, new_endpoints):
    """Create endpoint groups and their corresponding endpoints."""
    md = sa.MetaData()
    engine = op.get_bind()
    sa.Table('vpn_endpoint_groups', md, autoload_with=engine)
    op.bulk_insert(md.tables['vpn_endpoint_groups'], new_groups)
    sa.Table('vpn_endpoints', md, autoload_with=engine)
    op.bulk_insert(md.tables['vpn_endpoints'], new_endpoints)


def _update_connections(connection_map):
    """Store the endpoint group IDs in the connections."""
    for conn_id, mapping in connection_map.items():
        stmt = ipsec_site_conns.update().where(
            ipsec_site_conns.c.id == conn_id).values(
                local_ep_group_id=mapping['local'],
                peer_ep_group_id=mapping['peer'])
        op.execute(stmt)


def upgrade():
    new_groups = []
    new_endpoints = []
    service_map = {}
    session = sa.orm.Session(bind=op.get_bind())
    vpn_services = session.query(vpnservices).filter(
        vpnservices.c.subnet_id is not None).all()
    for vpn_service in vpn_services:
        subnet_id = vpn_service.subnet_id
        if subnet_id is None:
            continue  # Skip new service entries
        # Define the subnet group
        group_id = uuidutils.generate_uuid()
        group = {'id': group_id,
                 'name': '',
                 'description': '',
                 'tenant_id': vpn_service.tenant_id,
                 'endpoint_type': v_constants.SUBNET_ENDPOINT}
        new_groups.append(group)
        # Define the (sole) endpoint
        endpoint = {'endpoint_group_id': group_id,
                    'endpoint': subnet_id}
        new_endpoints.append(endpoint)
        # Save info to use for connections
        service_map[vpn_service.id] = group_id

    connection_map = {}
    ipsec_conns = session.query(ipsec_site_conns).all()
    for connection in ipsec_conns:
        peer_cidrs = session.query(ipsecpeercidrs.c.cidr).filter(
            ipsecpeercidrs.c.ipsec_site_connection_id == connection.id).all()
        if not peer_cidrs:
            continue  # Skip new style connections
        # Define the CIDR group
        group_id = uuidutils.generate_uuid()
        group = {'id': group_id,
                 'name': '',
                 'description': '',
                 'tenant_id': connection.tenant_id,
                 'endpoint_type': v_constants.CIDR_ENDPOINT}
        new_groups.append(group)
        # Define the endpoint(s)
        for peer_cidr in peer_cidrs:
            endpoint = {'endpoint_group_id': group_id,
                        'endpoint': peer_cidr[0]}
            new_endpoints.append(endpoint)
        # Save the endpoint group ID info for the connection
        vpn_service = connection.vpnservice_id
        connection_map[connection.id] = {'local': service_map[vpn_service],
                                         'peer': group_id}

    # Create all the defined endpoint groups and their endpoints
    _make_endpoint_groups(new_groups, new_endpoints)
    # Refer to new groups, in the IPSec connections
    _update_connections(connection_map)

    # Remove the peer_cidrs from IPSec connections
    op.execute(sa_expr.table('ipsecpeercidrs').delete())
    # Remove the subnets from VPN services
    stmt = vpnservices.update().where(
        vpnservices.c.subnet_id is not None).values(
            subnet_id=None)
    op.execute(stmt)
    session.commit()



# Copyright 2013, Nachi Ueno, NTT I3, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from neutron_lib.plugins import constants
from neutron_lib.plugins import directory
from neutron_lib import rpc as n_rpc
from oslo_log import log as logging
import oslo_messaging

from neutron_vpnaas.services.vpn.service_drivers import driver_validator

LOG = logging.getLogger(__name__)


class VpnDriver(metaclass=abc.ABCMeta):

    def __init__(self, service_plugin, validator=None):
        self.service_plugin = service_plugin
        if validator is None:
            validator = driver_validator.VpnDriverValidator(self)
        self.validator = validator
        self.name = ''

    @property
    def l3_plugin(self):
        return directory.get_plugin(constants.L3)

    @property
    def service_type(self):
        pass

    @abc.abstractmethod
    def create_vpnservice(self, context, vpnservice):
        pass

    @abc.abstractmethod
    def update_vpnservice(
        self, context, old_vpnservice, vpnservice):
        pass

    @abc.abstractmethod
    def delete_vpnservice(self, context, vpnservice):
        pass

    @abc.abstractmethod
    def create_ipsec_site_connection(self, context, ipsec_site_connection):
        pass

    @abc.abstractmethod
    def update_ipsec_site_connection(self, context, old_ipsec_site_connection,
                                     ipsec_site_connection):
        pass

    @abc.abstractmethod
    def delete_ipsec_site_connection(self, context, ipsec_site_connection):
        pass


class BaseIPsecVpnAgentApi:
    """Base class for IPSec API to agent."""

    def __init__(self, topic, default_version, driver):
        self.topic = topic
        self.driver = driver
        target = oslo_messaging.Target(topic=topic, version=default_version)
        self.client = n_rpc.get_client(target)

    def _agent_notification(self, context, method, router_id,
                            version=None, **kwargs):
        """Notify update for the agent.

        This method will find where is the router, and
        dispatch notification for the agent.
        """
        admin_context = context if context.is_admin else context.elevated()
        if not version:
            version = self.target.version
        l3_agents = self.driver.l3_plugin.get_l3_agents_hosting_routers(
            admin_context, [router_id],
            admin_state_up=True,
            active=True)
        for l3_agent in l3_agents:
            LOG.debug('Notify agent at %(topic)s.%(host)s the message '
                      '%(method)s %(args)s',
                      {'topic': self.topic,
                       'host': l3_agent.host,
                       'method': method,
                       'args': kwargs})
            cctxt = self.client.prepare(server=l3_agent.host, version=version)
            cctxt.cast(context, method, **kwargs)

    def vpnservice_updated(self, context, router_id, **kwargs):
        """Send update event of vpnservices."""
        kwargs['router'] = {'id': router_id}
        self._agent_notification(context, 'vpnservice_updated', router_id,
                                 **kwargs)



# Copyright 2017 Eayun, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


class VpnDriverValidator:
    """Driver-specific validation routines for VPN resources."""

    def __init__(self, driver):
        self.driver = driver

    @property
    def l3_plugin(self):
        return self.driver.l3_plugin

    def validate_ipsec_site_connection(self, context, ipsec_sitecon):
        """Driver can override this for its additional validations."""
        pass



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from rally.common import logging
from rally.task import scenario
from rally.task import types

import vpn_base

LOG = logging.getLogger(__name__)


class TestVpnStatusScenario(vpn_base.VpnBase):
    @types.convert(image={"type": "glance_image"},
                   flavor={"type": "nova_flavor"})
    @scenario.configure()
    def check_vpn_status(self, **kwargs):
        """Test VPN's status correctly after bringing router's status to
         DOWN and back to ACTIVE state

        1. Create 2 private networks, subnets and routers
        2. Create public network, subnets and GW IPs on routers, if not present
        3. Execute ip netns command and get the snat and qrouter namespaces
           (assuming we use DVR)
        4. Verify that there is a route between the router gateways by pinging
           each other from their snat namespaces
        5. Add security group rules for SSH and ICMP
        6. Start a nova instance in each of the private networks
        7. Create IKE and IPSEC policies
        8. Create VPN service at each of the routers
        9. Create IPSEC site connections at both endpoints
        10. Bring both the private router's status to DOWN state
        11. Verify that vpn-service and ipsec-site-connection is DOWN
        12. Bring back the router's status to ACTIVE state
        13. Verify the vpn-service and ipsec-site-connection is back to ACTIVE
        14. Perform resource cleanup
        """

        try:
            self.setup(**kwargs)
            self.create_networks(**kwargs)
            self.check_route()
            self.ike_policy = self._create_ike_policy(**kwargs)
            self.ipsec_policy = self._create_ipsec_policy(**kwargs)
            self.create_vpn_services()
            self.create_ipsec_site_connections(**kwargs)
            self.assert_statuses(final_status='ACTIVE', **kwargs)
            self.update_router(self.router_ids[0], admin_state_up=False)
            self.update_router(self.router_ids[1], admin_state_up=False)
            self.assert_statuses(final_status='DOWN', **kwargs)
            self.update_router(self.router_ids[0], admin_state_up=True)
            self.update_router(self.router_ids[1], admin_state_up=True)
            self.assert_statuses(final_status='ACTIVE', **kwargs)
            LOG.info("VPN STATUS TEST PASSED!")

        finally:
            self.cleanup()



# Copyright 2017 Eayun, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from neutron_lib.api.definitions import vpn_flavors
from neutron_lib.api import extensions
from neutron_lib import exceptions as nexception

from neutron_vpnaas._i18n import _


class FlavorsPluginNotLoaded(nexception.NotFound):
    message = _("Flavors plugin not found")


class NoProviderFoundForFlavor(nexception.NotFound):
    message = _("No service provider found for flavor %(flavor_id)s")


class Vpn_flavors(extensions.APIExtensionDescriptor):
    """Extension class supporting flavors for vpnservices."""
    api_definition = vpn_flavors



# Copyright (c) 2015 Canonical, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from oslo_config import cfg
from oslo_log import log as logging

from neutron.agent.linux import ip_lib
from neutron.agent.linux import utils
from neutron_lib import constants

from neutron_vpnaas._i18n import _
from neutron_vpnaas.services.vpn.device_drivers import ipsec

LOG = logging.getLogger(__name__)
TEMPLATE_PATH = os.path.dirname(os.path.abspath(__file__))

strongswan_opts = [
    cfg.StrOpt(
        'ipsec_config_template',
        default=os.path.join(
            TEMPLATE_PATH,
            'template/strongswan/ipsec.conf.template'),
        help=_('Template file for ipsec configuration.')),
    cfg.StrOpt(
        'strongswan_config_template',
        default=os.path.join(
            TEMPLATE_PATH,
            'template/strongswan/strongswan.conf.template'),
        help=_('Template file for strongswan configuration.')),
    cfg.StrOpt(
        'ipsec_secret_template',
        default=os.path.join(
            TEMPLATE_PATH,
            'template/strongswan/ipsec.secret.template'),
        help=_('Template file for ipsec secret configuration.')),
    cfg.StrOpt(
        'default_config_area',
        default=os.path.join(
            TEMPLATE_PATH,
            '/etc/strongswan.d'),
        help=_('The area where default StrongSwan configuration '
               'files are located.'))
]
cfg.CONF.register_opts(strongswan_opts, 'strongswan')


class StrongSwanProcess(ipsec.BaseSwanProcess):

    # ROUTED means route created. (only for auto=route mode)
    # CONNECTING means route created, connection tunnel is negotiating.
    # INSTALLED means route created,
    #           also connection tunnel installed. (traffic can pass)

    DIALECT_MAP = dict(ipsec.BaseSwanProcess.DIALECT_MAP)

    STATUS_DICT = {
        'ROUTED': constants.DOWN,
        'CONNECTING': constants.DOWN,
        'INSTALLED': constants.ACTIVE
    }
    STATUS_RE = r'([a-f0-9\-]+).* (ROUTED|CONNECTING|INSTALLED)'
    STATUS_NOT_RUNNING_RE = 'Command:.*ipsec.*status.*Exit code: [1|3] '

    def __init__(self, conf, process_id, vpnservice, namespace):
        dialect_map_update = {
            'v1': 'ikev1',
            'v2': 'ikev2',
            # ENCR_AES_CTR
            'aes-128-ctr': 'aes128ctr',
            'aes-192-ctr': 'aes192ctr',
            'aes-256-ctr': 'aes256ctr',
            # ENCR_AES_CCM_8
            'aes-128-ccm-8': 'aes128ccm8',
            'aes-192-ccm-8': 'aes192ccm8',
            'aes-256-ccm-8': 'aes256ccm8',
            # ENCR_AES_CCM_12
            'aes-128-ccm-12': 'aes128ccm12',
            'aes-192-ccm-12': 'aes192ccm12',
            'aes-256-ccm-12': 'aes256ccm12',
            # ENCR_AES_CCM_16
            'aes-128-ccm-16': 'aes128ccm16',
            'aes-192-ccm-16': 'aes192ccm16',
            'aes-256-ccm-16': 'aes256ccm16',
            # ENCR_AES_GCM_8
            'aes-128-gcm-8': 'aes128gcm8',
            'aes-192-gcm-8': 'aes192gcm8',
            'aes-256-gcm-8': 'aes256gcm8',
            # ENCR_AES_GCM_12
            'aes-128-gcm-12': 'aes128gcm12',
            'aes-192-gcm-12': 'aes192gcm12',
            'aes-256-gcm-12': 'aes256gcm12',
            # ENCR_AES_GCM_16
            'aes-128-gcm-16': 'aes128gcm16',
            'aes-192-gcm-16': 'aes192gcm16',
            'aes-256-gcm-16': 'aes256gcm16',
            # AUTH
            'sha256': 'sha256',
            'aes-xcbc': 'aesxcbc',
            'aes-cmac': 'aescmac',
            # PFS
            'group22': 'modp1024s160',
            'group23': 'modp2048s224',
            'group24': 'modp2048s256',
            'group25': 'ecp192',
            'group26': 'ecp224',
            'group27': 'ecp224bp',
            'group28': 'ecp256bp',
            'group29': 'ecp384bp',
            'group30': 'ecp512bp',
        }
        self.DIALECT_MAP.update(dialect_map_update)
        self._strongswan_piddir = self._get_strongswan_piddir()
        self._rootwrap_cfg = self._get_rootwrap_config()
        LOG.debug("strongswan piddir is '%s'", (self._strongswan_piddir))
        super().__init__(conf, process_id, vpnservice, namespace)

    def _get_strongswan_piddir(self):
        return utils.execute(
            cmd=[self.binary, "--piddir"], run_as_root=True).strip()

    def _check_status_line(self, line):
        """Parse a line and search for status information.

        If a given line contains status information for a connection,
        extract the status and mark the connection as ACTIVE or DOWN
        according to the STATUS_MAP.
        """
        m = self.STATUS_PATTERN.search(line)
        if m:
            connection_id = m.group(1)
            status = self.STATUS_MAP[m.group(2)]
            return connection_id, status
        return None, None

    def _execute(self, cmd, check_exit_code=True, extra_ok_codes=None):
        """Execute command on namespace.

        This execute is wrapped by namespace wrapper.
        The namespace wrapper will bind /etc/ and /var/run
        """
        ip_wrapper = ip_lib.IPWrapper(namespace=self.namespace)
        ns_wrapper = self.get_ns_wrapper()
        return ip_wrapper.netns.execute(
            [ns_wrapper,
             '--mount_paths=/etc:{}/etc,{}:{}/var/run'.format(
                 self.config_dir, self._strongswan_piddir, self.config_dir),
             ('--rootwrap_config=%s' % self._rootwrap_cfg
                 if self._rootwrap_cfg else ''),
             '--cmd=%s' % ','.join(cmd)],
            check_exit_code=check_exit_code,
            extra_ok_codes=extra_ok_codes)

    def copy_and_overwrite(self, from_path, to_path):
        # NOTE(toabctl): the agent may run as non-root user, so rm/copy as root
        if os.path.exists(to_path):
            utils.execute(
                cmd=["rm", "-rf", to_path], run_as_root=True)
        utils.execute(
            cmd=["cp", "-a", from_path, to_path], run_as_root=True)

    def ensure_configs(self):
        """Generate config files which are needed for StrongSwan.

        If there is no directory, this function will create
        dirs.
        """
        self.ensure_config_dir(self.vpnservice)
        self.ensure_config_file(
            'ipsec.conf',
            cfg.CONF.strongswan.ipsec_config_template,
            self.vpnservice)
        self.ensure_config_file(
            'strongswan.conf',
            cfg.CONF.strongswan.strongswan_config_template,
            self.vpnservice)
        self.ensure_config_file(
            'ipsec.secrets',
            cfg.CONF.strongswan.ipsec_secret_template,
            self.vpnservice,
            0o600)
        self.copy_and_overwrite(cfg.CONF.strongswan.default_config_area,
                                self._get_config_filename('strongswan.d'))

    def get_status(self):
        return self._execute([self.binary, 'status'],
                             extra_ok_codes=[1, 3])

    def restart(self):
        """Restart the process."""
        self.reload_secrets()
        self.reload()

    def reload_secrets(self):
        """Reload the ipsec.secrets file.

        Flushes and rereads all secrets defined in ipsec.secrets. This needs
        to be done each time when a new site connection is associated with
        a VPN service which already hosts a site connection - 'ipsec reload'
        does not reload the secrets and new connections will not authenticate
        properly.
        """
        self._execute([self.binary, 'rereadsecrets'])

    def reload(self):
        """Reload the process.

        Sends a USR1 signal to ipsec starter which in turn reloads the whole
        configuration on the running IKE daemon charon based on the actual
        ipsec.conf. Currently established connections are not affected by
        configuration changes.
        """
        self._execute([self.binary, 'reload'])

    def start(self):
        """Start the process for only auto=route mode now.

        Note: if there is no namespace yet,
        just do nothing, and wait next event.
        """
        if not self.namespace:
            return
        self._execute([self.binary, 'start'])
        # initiate ipsec connection
        for ipsec_site_conn in self.vpnservice['ipsec_site_connections']:
            self._execute([self.binary, 'stroke', 'up-nb',
                           ipsec_site_conn['id']])

    def stop(self):
        self._execute([self.binary, 'stop'])
        self.connection_status = {}


class StrongSwanDriver(ipsec.IPsecDriver):

    def create_process(self, process_id, vpnservice, namespace):
        return StrongSwanProcess(
            self.conf,
            process_id,
            vpnservice,
            namespace)



#    (c) Copyright 2013 Hewlett-Packard Development Company, L.P.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from neutron_lib.api.definitions import vpn
from neutron_lib.api import extensions
from neutron_lib import exceptions as nexception
from neutron_lib.plugins import constants as nconstants
from neutron_lib.services import base as service_base

from neutron.api.v2 import resource_helper

from neutron_vpnaas._i18n import _


class RouteInUseByVPN(nexception.InUse):
    """Operational error indicating a route is used for VPN.

    :param destinations: Destination CIDRs that are peers for VPN
    """
    message = _("Route(s) to %(destinations)s are used for VPN")


class VPNGatewayNotReady(nexception.BadRequest):
    message = _("VPN gateway not ready")


class VPNGatewayInError(nexception.Conflict):
    message = _("VPN gateway is in ERROR state. "
                "Please remove all errored VPN services and try again.")


class NoVPNAgentAvailable(nexception.ServiceUnavailable):
    message = _("No VPN agent available")


class Vpnaas(extensions.APIExtensionDescriptor):
    api_definition = vpn

    @classmethod
    def get_resources(cls):
        special_mappings = {'ikepolicies': 'ikepolicy',
                            'ipsecpolicies': 'ipsecpolicy'}
        plural_mappings = resource_helper.build_plural_mappings(
            special_mappings, vpn.RESOURCE_ATTRIBUTE_MAP)
        plural_mappings['peer_cidrs'] = 'peer_cidr'
        return resource_helper.build_resource_info(
            plural_mappings,
            vpn.RESOURCE_ATTRIBUTE_MAP,
            nconstants.VPN,
            register_quota=True,
            translate_name=True)

    @classmethod
    def get_plugin_interface(cls):
        return VPNPluginBase


class VPNPluginBase(service_base.ServicePluginBase, metaclass=abc.ABCMeta):

    def get_plugin_type(self):
        return nconstants.VPN

    def get_plugin_description(self):
        return 'VPN service plugin'

    @abc.abstractmethod
    def get_vpnservices(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_vpnservice(self, context, vpnservice_id, fields=None):
        pass

    @abc.abstractmethod
    def create_vpnservice(self, context, vpnservice):
        pass

    @abc.abstractmethod
    def update_vpnservice(self, context, vpnservice_id, vpnservice):
        pass

    @abc.abstractmethod
    def delete_vpnservice(self, context, vpnservice_id):
        pass

    @abc.abstractmethod
    def get_ipsec_site_connections(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_ipsec_site_connection(self, context,
                                  ipsecsite_conn_id, fields=None):
        pass

    @abc.abstractmethod
    def create_ipsec_site_connection(self, context, ipsec_site_connection):
        pass

    @abc.abstractmethod
    def update_ipsec_site_connection(self, context,
                                     ipsecsite_conn_id, ipsec_site_connection):
        pass

    @abc.abstractmethod
    def delete_ipsec_site_connection(self, context, ipsecsite_conn_id):
        pass

    @abc.abstractmethod
    def get_ikepolicy(self, context, ikepolicy_id, fields=None):
        pass

    @abc.abstractmethod
    def get_ikepolicies(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def create_ikepolicy(self, context, ikepolicy):
        pass

    @abc.abstractmethod
    def update_ikepolicy(self, context, ikepolicy_id, ikepolicy):
        pass

    @abc.abstractmethod
    def delete_ikepolicy(self, context, ikepolicy_id):
        pass

    @abc.abstractmethod
    def get_ipsecpolicies(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def get_ipsecpolicy(self, context, ipsecpolicy_id, fields=None):
        pass

    @abc.abstractmethod
    def create_ipsecpolicy(self, context, ipsecpolicy):
        pass

    @abc.abstractmethod
    def update_ipsecpolicy(self, context, ipsecpolicy_id, ipsecpolicy):
        pass

    @abc.abstractmethod
    def delete_ipsecpolicy(self, context, ipsecpolicy_id):
        pass



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import socket

import netaddr
from neutron.db import l3_db
from neutron.db import models_v2
from neutron_lib.api import validators
from neutron_lib import exceptions as nexception
from neutron_lib.exceptions import vpn as vpn_exception
from neutron_lib.plugins import constants as plugin_const
from neutron_lib.plugins import directory

from neutron_vpnaas._i18n import _
from neutron_vpnaas.services.vpn.common import constants


class VpnReferenceValidator:

    """
    Baseline validation routines for VPN resources.
    The validations here should be common to all VPN service providers and
    only raise exceptions from neutron_vpnaas.extensions.vpnaas.
    """

    IP_MIN_MTU = {4: 68, 6: 1280}

    @property
    def l3_plugin(self):
        try:
            return self._l3_plugin
        except AttributeError:
            self._l3_plugin = directory.get_plugin(plugin_const.L3)
            return self._l3_plugin

    @property
    def core_plugin(self):
        try:
            return self._core_plugin
        except AttributeError:
            self._core_plugin = directory.get_plugin()
            return self._core_plugin

    def _check_dpd(self, ipsec_sitecon):
        """Ensure that DPD timeout is greater than DPD interval."""
        if ipsec_sitecon['dpd_timeout'] <= ipsec_sitecon['dpd_interval']:
            raise vpn_exception.IPsecSiteConnectionDpdIntervalValueError(
                attr='dpd_timeout')

    def _check_mtu(self, context, mtu, ip_version):
        if mtu < VpnReferenceValidator.IP_MIN_MTU[ip_version]:
            raise vpn_exception.IPsecSiteConnectionMtuError(
                mtu=mtu, version=ip_version)

    def _validate_peer_address(self, ip_version, router):
        # NOTE: peer_address ip version should match with
        # at least one external gateway address ip version.
        # ipsec won't work with IPv6 LLA and neutron unaware GUA.
        # So to support vpnaas with ipv6, external network must
        # have ipv6 subnet
        for fixed_ip in router.gw_port['fixed_ips']:
            addr = fixed_ip['ip_address']
            if ip_version == netaddr.IPAddress(addr).version:
                return

        raise vpn_exception.ExternalNetworkHasNoSubnet(
            router_id=router.id,
            ip_version="IPv6" if ip_version == 6 else "IPv4")

    def resolve_peer_address(self, ipsec_sitecon, router):
        address = ipsec_sitecon['peer_address']
        # check if address is an ip address or fqdn
        invalid_ip_address = validators.validate_ip_address(address)
        if invalid_ip_address:
            # resolve fqdn
            try:
                addrinfo = socket.getaddrinfo(address, None)[0]
                ipsec_sitecon['peer_address'] = addrinfo[-1][0]
            except socket.gaierror:
                raise vpn_exception.VPNPeerAddressNotResolved(
                    peer_address=address)

        ip_version = netaddr.IPAddress(ipsec_sitecon['peer_address']).version
        self._validate_peer_address(ip_version, router)

    def _get_local_subnets(self, context, endpoint_group):
        if endpoint_group['type'] != constants.SUBNET_ENDPOINT:
            raise vpn_exception.WrongEndpointGroupType(
                group_type=endpoint_group['type'], which=endpoint_group['id'],
                expected=constants.SUBNET_ENDPOINT)
        subnet_ids = endpoint_group['endpoints']
        return context.session.query(models_v2.Subnet).filter(
            models_v2.Subnet.id.in_(subnet_ids)).all()

    def _get_peer_cidrs(self, endpoint_group):
        if endpoint_group['type'] != constants.CIDR_ENDPOINT:
            raise vpn_exception.WrongEndpointGroupType(
                group_type=endpoint_group['type'], which=endpoint_group['id'],
                expected=constants.CIDR_ENDPOINT)
        return endpoint_group['endpoints']

    def _check_local_endpoint_ip_versions(self, group_id, local_subnets):
        """Ensure all subnets in endpoint group have the same IP version.

        Will return the IP version, so it can be used for inter-group testing.
        """
        if len(local_subnets) == 1:
            return local_subnets[0]['ip_version']
        ip_versions = {subnet['ip_version'] for subnet in local_subnets}
        if len(ip_versions) > 1:
            raise vpn_exception.MixedIPVersionsForIPSecEndpoints(
                group=group_id)
        return ip_versions.pop()

    def _check_peer_endpoint_ip_versions(self, group_id, peer_cidrs):
        """Ensure all CIDRs in endpoint group have the same IP version.

        Will return the IP version, so it can be used for inter-group testing.
        """
        if len(peer_cidrs) == 1:
            return netaddr.IPNetwork(peer_cidrs[0]).version
        ip_versions = {netaddr.IPNetwork(pc).version for pc in peer_cidrs}
        if len(ip_versions) > 1:
            raise vpn_exception.MixedIPVersionsForIPSecEndpoints(
                group=group_id)
        return ip_versions.pop()

    def _check_peer_cidrs(self, peer_cidrs):
        """Ensure all CIDRs have the valid format."""
        for peer_cidr in peer_cidrs:
            msg = validators.validate_subnet(peer_cidr)
            if msg:
                raise vpn_exception.IPsecSiteConnectionPeerCidrError(
                    peer_cidr=peer_cidr)

    def _check_peer_cidrs_ip_versions(self, peer_cidrs):
        """Ensure all CIDRs have the same IP version."""
        if len(peer_cidrs) == 1:
            return netaddr.IPNetwork(peer_cidrs[0]).version
        ip_versions = {netaddr.IPNetwork(pc).version for pc in peer_cidrs}
        if len(ip_versions) > 1:
            raise vpn_exception.MixedIPVersionsForPeerCidrs()
        return ip_versions.pop()

    def _check_local_subnets_on_router(self, context, router, local_subnets):
        for subnet in local_subnets:
            self._check_subnet_id(context, router, subnet['id'])

    def _validate_compatible_ip_versions(self, local_ip_version,
                                         peer_ip_version):
        if local_ip_version != peer_ip_version:
            raise vpn_exception.MixedIPVersionsForIPSecConnection()

    def validate_ipsec_conn_optional_args(self, ipsec_sitecon, subnet):
        """Ensure that proper combinations of optional args are used.

        When VPN service has a subnet, then we must have peer_cidrs, and
        cannot have any endpoint groups. If no subnet for the service, then
        we must have both endpoint groups, and no peer_cidrs. Method will
        form a string indicating which endpoints are incorrect, for any
        exception raised.
        """

        local_epg_id = ipsec_sitecon.get('local_ep_group_id')
        peer_epg_id = ipsec_sitecon.get('peer_ep_group_id')
        peer_cidrs = ipsec_sitecon.get('peer_cidrs')
        if subnet:
            if not peer_cidrs:
                raise vpn_exception.MissingPeerCidrs()
            epgs = []
            if local_epg_id:
                epgs.append('local')
            if peer_epg_id:
                epgs.append('peer')
            if epgs:
                which = ' and '.join(epgs)
                suffix = 's' if len(epgs) > 1 else ''
                raise vpn_exception.InvalidEndpointGroup(which=which,
                                                         suffix=suffix)
        else:
            if peer_cidrs:
                raise vpn_exception.PeerCidrsInvalid()
            epgs = []
            if not local_epg_id:
                epgs.append('local')
            if not peer_epg_id:
                epgs.append('peer')
            if epgs:
                which = ' and '.join(epgs)
                suffix = 's' if len(epgs) > 1 else ''
                raise vpn_exception.MissingRequiredEndpointGroup(
                    which=which, suffix=suffix)

    def assign_sensible_ipsec_sitecon_defaults(self, ipsec_sitecon,
                                               prev_conn=None):
        """Provide defaults for optional items, if missing.

        With endpoint groups capabilities, the peer_cidr (legacy mode)
        and endpoint group IDs (new mode), are optional. For updating,
        we need to provide the previous values for any missing values,
        so that we can detect if the update request is attempting to
        mix modes.

        Flatten the nested DPD information, and set default values for
        any missing information. For connection updates, the previous
        values will be used as defaults for any missing items.
        """

        if prev_conn:
            ipsec_sitecon.setdefault(
                'peer_cidrs', [pc['cidr'] for pc in prev_conn['peer_cidrs']])
            ipsec_sitecon.setdefault('local_ep_group_id',
                                     prev_conn['local_ep_group_id'])
            ipsec_sitecon.setdefault('peer_ep_group_id',
                                     prev_conn['peer_ep_group_id'])
        else:
            prev_conn = {'dpd_action': 'hold',
                         'dpd_interval': 30,
                         'dpd_timeout': 120}
        dpd = ipsec_sitecon.get('dpd', {})
        ipsec_sitecon['dpd_action'] = dpd.get('action',
                                              prev_conn['dpd_action'])
        ipsec_sitecon['dpd_interval'] = dpd.get('interval',
                                                prev_conn['dpd_interval'])
        ipsec_sitecon['dpd_timeout'] = dpd.get('timeout',
                                               prev_conn['dpd_timeout'])

    def validate_ipsec_site_connection(self, context, ipsec_sitecon,
                                       local_ip_version, vpnservice=None):
        """Reference implementation of validation for IPSec connection.

        This makes sure that IP versions are the same. For endpoint groups,
        we use the local subnet(s) IP versions, and peer CIDR(s) IP versions.
        For legacy mode, we use the (sole) subnet IP version, and the peer
        CIDR(s). All IP versions must be the same.

        This method also checks peer_cidrs format(legacy mode),
        MTU (based on the local IP version), and DPD settings.
        """
        if not local_ip_version:
            # Using endpoint groups
            local_subnets = self._get_local_subnets(
                context, ipsec_sitecon['local_epg_subnets'])
            self._check_local_subnets_on_router(
                context, vpnservice['router_id'], local_subnets)
            local_ip_version = self._check_local_endpoint_ip_versions(
                ipsec_sitecon['local_ep_group_id'], local_subnets)
            peer_cidrs = self._get_peer_cidrs(ipsec_sitecon['peer_epg_cidrs'])
            peer_ip_version = self._check_peer_endpoint_ip_versions(
                ipsec_sitecon['peer_ep_group_id'], peer_cidrs)
        else:
            self._check_peer_cidrs(ipsec_sitecon['peer_cidrs'])
            peer_ip_version = self._check_peer_cidrs_ip_versions(
                ipsec_sitecon['peer_cidrs'])
        self._validate_compatible_ip_versions(local_ip_version,
                                              peer_ip_version)

        self._check_dpd(ipsec_sitecon)
        mtu = ipsec_sitecon.get('mtu')
        if mtu:
            self._check_mtu(context, mtu, local_ip_version)

    def _check_router(self, context, router_id):
        router = self.l3_plugin.get_router(context, router_id)
        if not router.get(l3_db.EXTERNAL_GW_INFO):
            raise vpn_exception.RouterIsNotExternal(router_id=router_id)

    def _check_subnet_id(self, context, router_id, subnet_id):
        ports = self.core_plugin.get_ports(
            context,
            filters={
                'fixed_ips': {'subnet_id': [subnet_id]},
                'device_id': [router_id]})
        if not ports:
            raise vpn_exception.SubnetIsNotConnectedToRouter(
                subnet_id=subnet_id,
                router_id=router_id)

    def validate_vpnservice(self, context, vpnservice):
        self._check_router(context, vpnservice['router_id'])
        if vpnservice['subnet_id'] is not None:
            self._check_subnet_id(context, vpnservice['router_id'],
                                  vpnservice['subnet_id'])

    def validate_ipsec_policy(self, context, ipsec_policy):
        """Reference implementation of validation for IPSec Policy.

        Service driver can override and implement specific logic
        for IPSec Policy validation.
        """
        pass

    def _validate_cidrs(self, cidrs):
        """Ensure valid IPv4/6 CIDRs."""
        for cidr in cidrs:
            msg = validators.validate_subnet(cidr)
            if msg:
                raise vpn_exception.InvalidEndpointInEndpointGroup(
                    group_type=constants.CIDR_ENDPOINT, endpoint=cidr,
                    why=_("Invalid CIDR"))

    def _validate_subnets(self, context, subnet_ids):
        """Ensure UUIDs OK and subnets exist."""
        for subnet_id in subnet_ids:
            msg = validators.validate_uuid(subnet_id)
            if msg:
                raise vpn_exception.InvalidEndpointInEndpointGroup(
                    group_type=constants.SUBNET_ENDPOINT, endpoint=subnet_id,
                    why=_('Invalid UUID'))
            try:
                self.core_plugin.get_subnet(context, subnet_id)
            except nexception.SubnetNotFound:
                raise vpn_exception.NonExistingSubnetInEndpointGroup(
                    subnet=subnet_id)

    def validate_endpoint_group(self, context, endpoint_group):
        """Reference validator for endpoint group.

        Ensures that there is at least one endpoint, all the endpoints in the
        group are of the same type, and that the endpoints are "valid".
        Note: Only called for create, as endpoints cannot be changed.
        """
        endpoints = endpoint_group['endpoints']
        if not endpoints:
            raise vpn_exception.MissingEndpointForEndpointGroup(
                group=endpoint_group)
        group_type = endpoint_group['type']
        if group_type == constants.CIDR_ENDPOINT:
            self._validate_cidrs(endpoints)
        elif group_type == constants.SUBNET_ENDPOINT:
            self._validate_subnets(context, endpoints)

    def validate_ike_policy(self, context, ike_policy):
        """Reference implementation of validation for IKE Policy.

        Service driver can override and implement specific logic
        for IKE Policy validation.
        """
        pass



# Copyright 2013, Nachi Ueno, NTT I3, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import contextlib
from unittest import mock

from neutron.db import servicetype_db as st_db
from neutron.services.flavors import flavors_plugin
from neutron.tests.unit.db import test_agentschedulers_db
from neutron.tests.unit.extensions import test_agent as test_agent_ext_plugin

from neutron_lib import constants as lib_constants
from neutron_lib import context
from neutron_lib import exceptions as lib_exc
from neutron_lib.exceptions import flavors as flav_exc
from neutron_lib.plugins import constants as p_constants
from neutron_lib.plugins import directory
from oslo_utils import uuidutils

from neutron_vpnaas.extensions import vpn_flavors
from neutron_vpnaas.services.vpn import plugin as vpn_plugin
from neutron_vpnaas.services.vpn.service_drivers import driver_validator
from neutron_vpnaas.services.vpn.service_drivers import ipsec as ipsec_driver
from neutron_vpnaas.tests import base
from neutron_vpnaas.tests.unit.db.vpn import test_vpn_db as test_db_vpnaas

FAKE_HOST = test_agent_ext_plugin.L3_HOSTA
VPN_DRIVER_CLASS = 'neutron_vpnaas.services.vpn.plugin.VPNDriverPlugin'

IPSEC_SERVICE_DRIVER = ('neutron_vpnaas.services.vpn.service_drivers.'
                        'ipsec.IPsecVPNDriver')
DUMMY_IPSEC_SERVICE_DRIVER = ('neutron_vpnaas.tests.unit.dummy_ipsec.'
                              'DummyIPsecVPNDriver')

_uuid = uuidutils.generate_uuid


class TestVPNDriverPlugin(test_db_vpnaas.TestVpnaas,
                          test_agentschedulers_db.AgentSchedulerTestMixIn,
                          test_agent_ext_plugin.AgentDBTestMixIn):

    def setUp(self):
        driver_cls_p = mock.patch(
            'neutron_vpnaas.services.vpn.'
            'service_drivers.ipsec.IPsecVPNDriver')
        driver_cls = driver_cls_p.start()
        self.driver = mock.Mock()
        self.driver.service_type = ipsec_driver.IPSEC
        self.driver.validator = driver_validator.VpnDriverValidator(
            self.driver)
        driver_cls.return_value = self.driver
        super().setUp(vpnaas_plugin=VPN_DRIVER_CLASS)
        # Note: Context must be created after BaseTestCase.setUp() so that
        # config for policy is set.
        self.adminContext = context.get_admin_context()

    def test_create_ipsec_site_connection(self, **extras):
        super().test_create_ipsec_site_connection()
        self.driver.create_ipsec_site_connection.assert_called_once_with(
            mock.ANY, mock.ANY)
        self.driver.delete_ipsec_site_connection.assert_called_once_with(
            mock.ANY, mock.ANY)

    def test_create_vpnservice(self):
        mock.patch('neutron_vpnaas.services.vpn.plugin.'
                   'VPNDriverPlugin._get_driver_for_vpnservice',
                   return_value=self.driver).start()
        stm = directory.get_plugin(p_constants.VPN).service_type_manager
        stm.add_resource_association = mock.Mock()
        super().test_create_vpnservice()
        self.driver.create_vpnservice.assert_called_once_with(
            mock.ANY, mock.ANY)
        stm.add_resource_association.assert_called_once_with(
            mock.ANY, p_constants.VPN, 'vpnaas', mock.ANY)

    def test_delete_vpnservice(self, **extras):
        stm = directory.get_plugin(p_constants.VPN).service_type_manager
        stm.del_resource_associations = mock.Mock()
        super().test_delete_vpnservice()
        self.driver.delete_vpnservice.assert_called_once_with(
            mock.ANY, mock.ANY)
        stm.del_resource_associations.assert_called_once_with(
            mock.ANY, [mock.ANY])

    def test_update_vpnservice(self, **extras):
        super().test_update_vpnservice()
        self.driver.update_vpnservice.assert_called_once_with(
            mock.ANY, mock.ANY, mock.ANY)

    @contextlib.contextmanager
    def vpnservice_set(self):
        """Test case to create a ipsec_site_connection."""
        vpnservice_name = "vpn1"
        ipsec_site_connection_name = "ipsec_site_connection"
        ikename = "ikepolicy1"
        ipsecname = "ipsecpolicy1"
        description = "my-vpn-connection"
        keys = {'name': vpnservice_name,
                'description': "my-vpn-connection",
                'peer_address': '192.168.1.10',
                'peer_id': '192.168.1.10',
                'peer_cidrs': ['192.168.2.0/24', '192.168.3.0/24'],
                'initiator': 'bi-directional',
                'mtu': 1500,
                'dpd_action': 'hold',
                'dpd_interval': 40,
                'dpd_timeout': 120,
                'tenant_id': self._tenant_id,
                'psk': 'abcd',
                'status': 'PENDING_CREATE',
                'admin_state_up': True}
        with self.ikepolicy(name=ikename) as ikepolicy:
            with self.ipsecpolicy(name=ipsecname) as ipsecpolicy:
                with self.subnet() as subnet:
                    with self.router() as router:
                        plugin = directory.get_plugin()
                        agent = {'host': FAKE_HOST,
                                 'agent_type': lib_constants.AGENT_TYPE_L3,
                                 'binary': 'fake-binary',
                                 'topic': 'fake-topic'}
                        plugin.create_or_update_agent(self.adminContext, agent)
                        plugin.schedule_router(
                            self.adminContext, router['router']['id'])
                        with self.vpnservice(name=vpnservice_name,
                                             subnet=subnet,
                                             router=router) as vpnservice1:
                            keys['ikepolicy_id'] = ikepolicy['ikepolicy']['id']
                            keys['ipsecpolicy_id'] = (
                                ipsecpolicy['ipsecpolicy']['id']
                            )
                            keys['vpnservice_id'] = (
                                vpnservice1['vpnservice']['id']
                            )
                            with self.ipsec_site_connection(
                                self.fmt,
                                ipsec_site_connection_name,
                                keys['peer_address'],
                                keys['peer_id'],
                                keys['peer_cidrs'],
                                keys['mtu'],
                                keys['psk'],
                                keys['initiator'],
                                keys['dpd_action'],
                                keys['dpd_interval'],
                                keys['dpd_timeout'],
                                vpnservice1,
                                ikepolicy,
                                ipsecpolicy,
                                keys['admin_state_up'],
                                description=description,
                            ):
                                yield vpnservice1['vpnservice']

    def test_update_status(self):
        with self.vpnservice_set() as vpnservice:
            self._register_agent_states()
            service_plugin = directory.get_plugin(p_constants.VPN)
            service_plugin.update_status_by_agent(
                self.adminContext,
                [{'status': 'ACTIVE',
                  'ipsec_site_connections': {},
                  'updated_pending_status': True,
                  'id': vpnservice['id']}])
            vpnservice = service_plugin.get_vpnservice(
                self.adminContext, vpnservice['id'])
            self.assertEqual(lib_constants.ACTIVE, vpnservice['status'])


class TestVPNDriverPluginMultipleDrivers(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        vpnaas_providers = [
            {'service_type': p_constants.VPN,
             'name': 'ipsec',
             'driver': IPSEC_SERVICE_DRIVER,
             'default': True},
            {'service_type': p_constants.VPN,
             'name': 'dummy',
             'driver': DUMMY_IPSEC_SERVICE_DRIVER,
             'default': False}]
        self.service_providers = (
            mock.patch.object(st_db.ServiceTypeManager,
                              'get_service_providers').start())
        self.service_providers.return_value = vpnaas_providers
        self.adminContext = context.get_admin_context()

    @contextlib.contextmanager
    def vpnservices_providers_set(self, vpnservices=None, provider_names=None):
        if not vpnservices:
            vpnservices = []
        if not provider_names:
            provider_names = {}
        stm = st_db.ServiceTypeManager()
        stm.get_provider_names_by_resource_ids = mock.Mock(
            return_value=provider_names)
        mock.patch('neutron.db.servicetype_db.ServiceTypeManager.get_instance',
                   return_value=stm).start()
        mock.patch('neutron_vpnaas.db.vpn.vpn_db.VPNPluginDb.get_vpnservices',
                   return_value=vpnservices).start()
        yield stm

    def test_multiple_drivers_loaded(self):
        with self.vpnservices_providers_set():
            driver_plugin = vpn_plugin.VPNDriverPlugin()
            self.assertEqual(2, len(driver_plugin.drivers))
            self.assertEqual('ipsec', driver_plugin.default_provider)
            self.assertIn('ipsec', driver_plugin.drivers)
            self.assertEqual('ipsec', driver_plugin.drivers['ipsec'].name)
            self.assertIn('dummy', driver_plugin.drivers)
            self.assertEqual('dummy', driver_plugin.drivers['dummy'].name)

    def test_provider_lost(self):
        LOST_SERVICE_ID = _uuid()
        LOST_PROVIDER_SERVICE = {'id': LOST_SERVICE_ID}
        with self.vpnservices_providers_set(
                vpnservices=[LOST_PROVIDER_SERVICE],
                provider_names={LOST_SERVICE_ID: 'LOST_PROVIDER'}
        ):
            self.assertRaises(SystemExit, vpn_plugin.VPNDriverPlugin)

    def test_unasso_vpnservices(self):
        UNASSO_SERVICE_ID = _uuid()
        with self.vpnservices_providers_set(
                vpnservices=[{'id': UNASSO_SERVICE_ID}]
        ) as stm:
            stm.add_resource_association = mock.Mock()
            vpn_plugin.VPNDriverPlugin()
            stm.add_resource_association.assert_called_once_with(
                mock.ANY, p_constants.VPN, 'ipsec', UNASSO_SERVICE_ID)

    def test_get_driver_for_vpnservice(self):
        DUMMY_VPNSERVICE_ID = _uuid()
        DUMMY_VPNSERVICE = {'id': DUMMY_VPNSERVICE_ID}
        provider_names = {DUMMY_VPNSERVICE_ID: 'dummy'}
        with self.vpnservices_providers_set(provider_names=provider_names):
            driver_plugin = vpn_plugin.VPNDriverPlugin()
            self.assertEqual(
                driver_plugin.drivers['dummy'],
                driver_plugin._get_driver_for_vpnservice(
                    self.adminContext, DUMMY_VPNSERVICE))

    def test_get_driver_for_ipsec_site_connection(self):
        IPSEC_VPNSERVICE_ID = _uuid()
        IPSEC_SITE_CONNECTION = {'vpnservice_id': IPSEC_VPNSERVICE_ID}
        provider_names = {IPSEC_VPNSERVICE_ID: 'ipsec'}
        with self.vpnservices_providers_set(provider_names=provider_names):
            driver_plugin = vpn_plugin.VPNDriverPlugin()
            self.assertEqual(
                driver_plugin.drivers['ipsec'],
                driver_plugin._get_driver_for_ipsec_site_connection(
                    self.adminContext, IPSEC_SITE_CONNECTION))

    def test_get_provider_for_none_flavor_id(self):
        with self.vpnservices_providers_set():
            driver_plugin = vpn_plugin.VPNDriverPlugin()
            provider = driver_plugin._get_provider_for_flavor(
                self.adminContext, None)
            self.assertEqual(
                driver_plugin.default_provider, provider)

    def test_get_provider_for_flavor_id_plugin_not_loaded(self):
        with self.vpnservices_providers_set():
            driver_plugin = vpn_plugin.VPNDriverPlugin()
            self.assertRaises(
                vpn_flavors.FlavorsPluginNotLoaded,
                driver_plugin._get_provider_for_flavor,
                self.adminContext,
                _uuid())

    def test_get_provider_for_flavor_id_invalid_type(self):
        FAKE_FLAVOR = {'service_type': 'NOT_VPN'}
        directory.add_plugin(p_constants.FLAVORS,
                             flavors_plugin.FlavorsPlugin())
        mock.patch(
            'neutron.services.flavors.flavors_plugin.FlavorsPlugin.get_flavor',
            return_value=FAKE_FLAVOR).start()
        with self.vpnservices_providers_set():
            driver_plugin = vpn_plugin.VPNDriverPlugin()
            self.assertRaises(
                lib_exc.InvalidServiceType,
                driver_plugin._get_provider_for_flavor,
                self.adminContext,
                _uuid())

    def test_get_provider_for_flavor_id_flavor_disabled(self):
        FAKE_FLAVOR = {'service_type': p_constants.VPN,
                       'enabled': False}
        directory.add_plugin(p_constants.FLAVORS,
                             flavors_plugin.FlavorsPlugin())
        mock.patch(
            'neutron.services.flavors.flavors_plugin.FlavorsPlugin.get_flavor',
            return_value=FAKE_FLAVOR).start()
        with self.vpnservices_providers_set():
            driver_plugin = vpn_plugin.VPNDriverPlugin()
            self.assertRaises(
                flav_exc.FlavorDisabled,
                driver_plugin._get_provider_for_flavor,
                self.adminContext,
                _uuid())

    def test_get_provider_for_flavor_id_provider_not_found(self):
        FLAVOR_ID = _uuid()
        FAKE_FLAVOR = {'id': FLAVOR_ID,
                       'service_type': p_constants.VPN,
                       'enabled': True}
        PROVIDERS = [{'provider': 'SOME_PROVIDER'}]
        directory.add_plugin(p_constants.FLAVORS,
                             flavors_plugin.FlavorsPlugin())
        mock.patch(
            'neutron.services.flavors.flavors_plugin.FlavorsPlugin.get_flavor',
            return_value=FAKE_FLAVOR).start()
        mock.patch(
            'neutron.services.flavors.flavors_plugin.'
            'FlavorsPlugin.get_flavor_next_provider',
            return_value=PROVIDERS).start()
        with self.vpnservices_providers_set():
            driver_plugin = vpn_plugin.VPNDriverPlugin()
            self.assertRaises(
                vpn_flavors.NoProviderFoundForFlavor,
                driver_plugin._get_provider_for_flavor,
                self.adminContext,
                FLAVOR_ID)

    def test_get_provider_for_flavor_id(self):
        FLAVOR_ID = _uuid()
        FAKE_FLAVOR = {'id': FLAVOR_ID,
                       'service_type': p_constants.VPN,
                       'enabled': True}
        PROVIDERS = [{'provider': 'dummy'}]
        directory.add_plugin(p_constants.FLAVORS,
                             flavors_plugin.FlavorsPlugin())
        mock.patch(
            'neutron.services.flavors.flavors_plugin.FlavorsPlugin.get_flavor',
            return_value=FAKE_FLAVOR).start()
        mock.patch(
            'neutron.services.flavors.flavors_plugin.'
            'FlavorsPlugin.get_flavor_next_provider',
            return_value=PROVIDERS).start()
        with self.vpnservices_providers_set():
            driver_plugin = vpn_plugin.VPNDriverPlugin()
            self.assertEqual(
                'dummy',
                driver_plugin._get_provider_for_flavor(
                    self.adminContext, FLAVOR_ID))



# Copyright 2015 Awcloud Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions as nexception

from neutron_vpnaas._i18n import _
from neutron_vpnaas.services.vpn.service_drivers import driver_validator


class IpsecValidationFailure(nexception.BadRequest):
    message = _("IPSec does not support %(resource)s attribute %(key)s "
                "with value '%(value)s'")


class IpsecVpnValidator(driver_validator.VpnDriverValidator):

    """Driver-specific validator methods for the Openswan, Strongswan
    and Libreswan.
    """

    def _check_transform_protocol(self, context, transform_protocol):
        """Restrict selecting ah-esp as IPSec Policy transform protocol.

        For those *Swan implementations, the 'ah-esp' transform protocol
        is not supported and therefore the request should be rejected.
        """
        if transform_protocol == "ah-esp":
            raise IpsecValidationFailure(
                resource='IPsec Policy',
                key='transform_protocol',
                value=transform_protocol)

    def validate_ipsec_policy(self, context, ipsec_policy):
        transform_protocol = ipsec_policy.get('transform_protocol')
        self._check_transform_protocol(context, transform_protocol)

    def validate_ipsec_site_connection(self, context, ipsec_sitecon):
        if 'ipsecpolicy_id' in ipsec_sitecon:
            ipsec_policy = self.driver.service_plugin.get_ipsecpolicy(
                context, ipsec_sitecon['ipsecpolicy_id'])
            self.validate_ipsec_policy(context, ipsec_policy)



# Copyright 2020 cmss, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op
import sqlalchemy as sa

"""add_aggressive_negotiation_modes

Revision ID: 5f884db48ba9
Revises: 95601446dbcc
Create Date: 2020-05-12 14:37:46.320070

"""

# revision identifiers, used by Alembic.
revision = '5f884db48ba9'
down_revision = '95601446dbcc'

phase1_negotiation_modes = sa.Enum('main', 'aggressive',
                                   name='ike_phase1_mode')


def upgrade():
    op.alter_column('ikepolicies', 'phase1_negotiation_mode',
                    type_=phase1_negotiation_modes,
                    existing_nullable=False)



# Copyright (c) 2015 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import os

from neutron.agent.linux import ip_lib
from neutron_vpnaas.services.vpn.device_drivers import ipsec


class LibreSwanProcess(ipsec.OpenSwanProcess):
    """Libreswan Process manager class.

    Libreswan needs nssdb initialised before running pluto daemon.
    """
    IPSEC_CONF_NAT_TRAVERSAL = None

    # pylint: disable=useless-super-delegation
    def __init__(self, conf, process_id, vpnservice, namespace):
        self._rootwrap_cfg = self._get_rootwrap_config()
        super().__init__(conf, process_id, vpnservice, namespace)

    def _ipsec_execute(self, cmd, check_exit_code=True, extra_ok_codes=None):
        """Execute ipsec command on namespace.

        This execute is wrapped by namespace wrapper.
        The namespace wrapper will bind /etc and /var/run
        """
        ip_wrapper = ip_lib.IPWrapper(namespace=self.namespace)
        mount_paths = {'/etc': '%s/etc' % self.config_dir,
                       '/run': '%s/var/run' % self.config_dir}
        mount_paths_str = ','.join(
            "{}:{}".format(source, target)
            for source, target in mount_paths.items())
        ns_wrapper = self.get_ns_wrapper()
        return ip_wrapper.netns.execute(
            [ns_wrapper,
             '--mount_paths=%s' % mount_paths_str,
             ('--rootwrap_config=%s' % self._rootwrap_cfg
                 if self._rootwrap_cfg else ''),
             '--cmd={},{}'.format(self.binary, ','.join(cmd))],
            check_exit_code=check_exit_code,
            extra_ok_codes=extra_ok_codes)

    def _ensure_needed_files(self):
        # addconn reads from /etc/hosts and /etc/resolv.conf. As /etc would be
        # bind-mounted, create these two empty files in the target directory.
        with open('%s/etc/hosts' % self.config_dir, 'a'):
            pass
        with open('%s/etc/resolv.conf' % self.config_dir, 'a'):
            pass

    def ensure_configs(self):
        """Generate config files which are needed for Libreswan.

        Initialise the nssdb, otherwise pluto daemon will fail to run.
        """

        # Since we set ipsec.secrets to be owned by root, the standard
        # mechanisms for setting up the config files will get a permission
        # problem when attempting to overwrite the file, so we need to
        # remove it first.
        secrets_file = self._get_config_filename('ipsec.secrets')
        if os.path.exists(secrets_file):
            self._execute(['rm', '-f', secrets_file])

        super().ensure_configs()

        # LibreSwan uses the capabilities library to restrict access to
        # ipsec.secrets to users that have explicit access. Since pluto is
        # running as root and the file has 0600 perms, we must set the
        # owner of the file to root.
        self._execute(['chown', '--from=%s' % os.getuid(), 'root:root',
                       secrets_file])

        # Libreswan needs to write logs to this directory.
        self._execute(['chown', '--from=%s' % os.getuid(), 'root:root',
                       self.log_dir])

        self._ensure_needed_files()

        # Load the ipsec kernel module if not loaded
        self._ipsec_execute(['_stackmanager', 'start'])
        # checknss creates nssdb only if it is missing
        # It is added in Libreswan version v3.10
        # For prior versions use initnss
        try:
            self._ipsec_execute(['checknss'])
        except RuntimeError:
            self._ipsec_execute(['initnss'])

    def get_status(self):
        return self._ipsec_execute(['whack', '--status'],
                                   extra_ok_codes=[1, 3])

    def start_pluto(self):
        cmd = ['pluto',
               '--use-xfrm',
               '--uniqueids']

        if self.conf.ipsec.enable_detailed_logging:
            cmd += ['--perpeerlog', '--perpeerlogbase', self.log_dir]
        self._ipsec_execute(cmd)

    def add_ipsec_connection(self, nexthop, conn_id):
        # Connections will be automatically added as auto=start/add for
        # initiator=bi-directional/response-only specified in the config.
        pass

    def start_whack_listening(self):
        # NOTE(huntxu): This is a workaround for with a weak (len<8) secret,
        # "ipsec whack --listen" will exit with 3.
        self._ipsec_execute(['whack', '--listen'], extra_ok_codes=[3])

    def shutdown_whack(self):
        self._ipsec_execute(['whack', '--shutdown'])

    def initiate_connection(self, conn_name):
        self._ipsec_execute(
            ['whack', '--name', conn_name, '--asynchronous', '--initiate'])

    def terminate_connection(self, conn_name):
        self._ipsec_execute(['whack', '--name', conn_name, '--terminate'])


class LibreSwanDriver(ipsec.IPsecDriver):
    def create_process(self, process_id, vpnservice, namespace):
        return LibreSwanProcess(
            self.conf,
            process_id,
            vpnservice,
            namespace)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""kilo

Revision ID: kilo
Revises: 3ea02b2a773e
Create Date: 2015-04-16 00:00:00.000000

"""

# revision identifiers, used by Alembic.
revision = 'kilo'
down_revision = '3ea02b2a773e'


def upgrade():
    """A no-op migration for marking the Kilo release."""
    pass



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import netaddr
from neutron.agent.linux import ip_lib
from neutron.common import config as common_config
from neutron.common.ovn import constants as ovn_const
from neutron.conf.agent import common as agent_conf
from neutron.conf import common as common_conf
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf
from neutron.conf.plugins.ml2.drivers import ovs_conf
from neutron.tests.common import net_helpers
from neutron.tests.functional import base
from neutron_lib import constants as lib_constants
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from neutron_lib.utils import helpers
from oslo_config import cfg
from ovsdbapp.backend.ovs_idl import event

from neutron_vpnaas.agent.ovn.vpn import agent
from neutron_vpnaas.agent.ovn.vpn import ovsdb
from neutron_vpnaas.services.vpn.common import constants as vpn_const
from neutron_vpnaas.services.vpn.device_drivers import ipsec
from neutron_vpnaas.services.vpn import ovn_agent
from neutron_vpnaas.services.vpn.service_drivers import ovn_ipsec


OVS_INTERFACE_DRIVER = 'neutron.agent.linux.interface.OVSInterfaceDriver'
IPSEC_SERVICE_PROVIDER = ('VPN:ovn:neutron_vpnaas.services.vpn.'
                          'service_drivers.ovn_ipsec.IPsecOvnVPNDriver:'
                          'default')
VPN_PLUGIN = 'neutron_vpnaas.services.vpn.ovn_plugin.VPNOVNDriverPlugin'

PUBLIC_NET = netaddr.IPNetwork('19.4.4.0/24')
LOCAL_NETS = list(netaddr.IPNetwork('10.0.0.0/16').subnet(24))
PEER_NET = netaddr.IPNetwork('10.1.0.0/16')
PEER_ADDR = '19.4.5.6'


class VPNAgentHealthEvent(event.WaitEvent):
    event_name = 'VPNAgentHealthEvent'

    def __init__(self, chassis, sb_cfg, table, timeout=5):
        self.chassis = chassis
        self.sb_cfg = sb_cfg
        super().__init__(
            (self.ROW_UPDATE,), table, (('name', '=', self.chassis),),
            timeout=timeout)

    def matches(self, event, row, old=None):
        if not super().matches(event, row, old):
            return False
        return int(row.external_ids.get(
            vpn_const.OVN_AGENT_VPN_SB_CFG_KEY, 0)) >= self.sb_cfg


class OvnSiteInfo:
    def __init__(self, parent, index, ext_net, ext_sub):
        self.ext_net = ext_net
        self.ext_sub = ext_sub
        self.parent = parent
        self.context = parent.context
        self.fmt = parent.fmt
        self.index = index

    def create_base(self):
        router_data = {
            'name': 'r%d' % self.index,
            'admin_state_up': True,
            'tenant_id': self.parent._tenant_id,
            'external_gateway_info': {
                'enable_snat': True,
                'network_id': self.ext_net['id'],
                'external_fixed_ips': [
                    {'ip_address': str(PUBLIC_NET[4 + 2 * self.index]),
                     'subnet_id': self.ext_sub['id']}
                ]
            }
        }
        self.router = self.parent.l3_plugin.create_router(
            self.context, {'router': router_data})

        # local subnet
        private_net = LOCAL_NETS[self.index]
        self.local_cidr = str(private_net)

        net = self.parent._make_network(self.fmt, 'local%d' % self.index, True)
        self.local_net = net['network']
        sub = self.parent._make_subnet(self.fmt, net, private_net[1],
                                       self.local_cidr, enable_dhcp=False)
        self.local_sub = sub['subnet']
        interface_info = {'subnet_id': self.local_sub['id']}
        self.parent.l3_plugin.add_router_interface(
            self.context, self.router['id'], interface_info)

    def create_vpnservice(self):
        plugin = self.parent.vpn_plugin
        data = {
            'tenant_id': self.parent._tenant_id,
            'name': 'my-service',
            'description': 'new service',
            'subnet_id': self.local_sub['id'],
            'router_id': self.router['id'],
            'flavor_id': None,
            'admin_state_up': True,
        }
        self.vpnservice = plugin.create_vpnservice(self.context,
                                                   {'vpnservice': data})
        self.local_addr = self.vpnservice['external_v4_ip']

        data = {
            'tenant_id': self.parent._tenant_id,
            'name': 'ikepolicy%d' % self.index,
            'description': '',
            'auth_algorithm': 'sha1',
            'encryption_algorithm': 'aes-128',
            'phase1_negotiation_mode': 'main',
            'ike_version': 'v1',
            'pfs': 'group5',
            'lifetime': {'units': 'seconds', 'value': 3600},
        }
        self.ikepolicy = plugin.create_ikepolicy(self.context,
                                                 {'ikepolicy': data})

        data = {
            'tenant_id': self.parent._tenant_id,
            'name': 'ipsecpolicy%d' % self.index,
            'description': '',
            'transform_protocol': 'esp',
            'auth_algorithm': 'sha1',
            'encryption_algorithm': 'aes-128',
            'encapsulation_mode': 'tunnel',
            'pfs': 'group5',
            'lifetime': {'units': 'seconds', 'value': 3600},
        }
        self.ipsecpolicy = plugin.create_ipsecpolicy(self.context,
                                                     {'ipsecpolicy': data})

    def create_site_connection(self, peer_addr, peer_cidr):
        data = {
            'tenant_id': self.parent._tenant_id,
            'name': 'conn%d' % self.index,
            'description': '',
            'local_id': self.local_addr,
            'peer_address': peer_addr,
            'peer_id': peer_addr,
            'peer_cidrs': [peer_cidr],
            'mtu': 1500,
            'initiator': 'bi-directional',
            'auth_mode': 'psk',
            'psk': 'secret',
            'dpd': {
                'action': 'hold',
                'interval': 30,
                'timeout': 120,
            },
            'admin_state_up': True,
            'vpnservice_id': self.vpnservice['id'],
            'ikepolicy_id': self.ikepolicy['id'],
            'ipsecpolicy_id': self.ipsecpolicy['id'],
            'local_ep_group_id': None,
            'peer_ep_group_id': None,
        }
        self.siteconn = self.parent.vpn_plugin.create_ipsec_site_connection(
            self.context, {'ipsec_site_connection': data})


class TestOvnVPNAgentBase(base.TestOVNFunctionalBase):
    FAKE_CHASSIS_HOST = 'ovn-host-fake'

    def setUp(self):
        cfg.CONF.set_override('service_provider', [IPSEC_SERVICE_PROVIDER],
                              group='service_providers')
        service_plugins = {'vpnaas_plugin': VPN_PLUGIN}
        super().setUp(service_plugins=service_plugins)
        common_config.register_common_config_options()

        self.mock_ovsdb_idl = mock.Mock()
        mock_instance = mock.Mock()
        mock_instance.start.return_value = self.mock_ovsdb_idl
        mock_ovs_idl = mock.patch.object(ovsdb, 'VPNAgentOvsIdl').start()
        mock_ovs_idl.return_value = mock_instance

        self.vpn_plugin = directory.get_plugin(plugin_constants.VPN)
        # normally called in post_for_initialize
        self.vpn_plugin.watch_agent_events()
        self.vpn_service_driver = self.vpn_plugin.drivers['ovn']

        self.handler = self.sb_api.idl.notify_handler
        self.agent = self._start_vpn_agent()
        self.agent_driver = self.agent.device_drivers[0]

    def _start_vpn_agent(self):
        # Set up a ConfigOpts separate to cfg.CONF in order to avoid conflicts
        # with other tests.
        # The OVN VPN agent registers a different variant of
        # vpnagent.vpn_device_drivers than the L3 agent extension.
        conf = agent_conf.setup_conf()
        conf.register_opts(ovn_conf.ovn_opts, group='ovn')
        conf.register_opts(ipsec.ipsec_opts, 'ipsec')
        common_conf.register_core_common_config_opts(conf)
        ovs_conf.register_ovs_opts(conf)
        ovn_agent.register_opts(conf)
        agent_conf.register_process_monitor_opts(conf)
        agent_conf.setup_privsep()

        conf.set_override('state_path', self.get_default_temp_dir().path)
        conf.set_override('interface_driver', OVS_INTERFACE_DRIVER)
        conf.set_override('vpn_device_driver', [self.VPN_DEVICE_DRIVER],
                          group='vpnagent')

        ovn_sb_db = self.ovsdb_server_mgr.get_ovsdb_connection_path('sb')
        conf.set_override('ovn_sb_connection', ovn_sb_db, group='ovn')

        self.chassis_name = self.add_fake_chassis(self.FAKE_CHASSIS_HOST)
        mock.patch.object(agent.OvnVpnAgent,
                          '_get_own_chassis_name',
                          return_value=self.chassis_name).start()
        conf.set_override('host', self.FAKE_CHASSIS_HOST)

        self.br_int = self.useFixture(net_helpers.OVSBridgeFixture()).bridge
        conf.set_override('integration_bridge', self.br_int.br_name, 'OVS')

        # name prefix for namespaces managed by vpn agent
        # will be patched into device driver to make sure concurrent
        # tests don't interfere with each other
        # (a vpn agent will normally remove all unknown qvpn- namespaces)
        self.ns_prefix = 'qvpn-test-%s-' % helpers.get_random_string(8)

        agt = agent.OvnVpnAgent(conf)
        driver = agt.device_drivers[0]
        driver.agent_rpc = mock.Mock()
        # let initial sync get an empty list of vpnservices
        driver.agent_rpc.get_vpn_services_on_host.return_value = []
        driver.devmgr.plugin = driver.agent_rpc
        driver.devmgr.OVN_NS_PREFIX = self.ns_prefix

        agt.start()
        self.addCleanup(agt.ovs_idl.ovsdb_connection.stop)
        self.addCleanup(agt.sb_idl.ovsdb_connection.stop)
        # let agent remove remaining vpn namespaces in cleanup
        self.addCleanup(driver._cleanup_stale_vpn_processes, [])

        return agt

    @property
    def agent_chassis_table(self):
        if self.agent.has_chassis_private:
            return 'Chassis_Private'
        return 'Chassis'

    def _make_ext_network(self):
        network = self._make_network(
            self.fmt, 'external-net', True, as_admin=True,
            arg_list=('router:external',
                      'provider:network_type',
                      'provider:physical_network'),
            **{'router:external': True,
                'provider:network_type': 'flat',
                'provider:physical_network': 'public'})

        pools = [{'start': PUBLIC_NET[2], 'end': PUBLIC_NET[253]}]
        gateway = PUBLIC_NET[1]
        cidr = str(PUBLIC_NET)
        subnet = self._make_subnet(self.fmt, network, gateway, cidr,
                                   allocation_pools=pools,
                                   enable_dhcp=False)
        return network['network'], subnet['subnet']

    def _find_lswitch_by_neutron_name(self, name):
        for row in self.nb_api._tables['Logical_Switch'].rows.values():
            if (row.external_ids.get(
                    ovn_const.OVN_NETWORK_NAME_EXT_ID_KEY) == name):
                return row

    def _find_transit_lswitch(self, router_id):
        name = ovn_ipsec.TRANSIT_NETWORK_PREFIX + router_id
        return self._find_lswitch_by_neutron_name(name)

    def _match_extids(self, row, expected):
        for key, value in expected.items():
            if row.external_ids.get(key) != value:
                return False
        return True

    def _find_transit_ns_port(self, router_id, ports):
        name = ovn_ipsec.TRANSIT_PORT_PREFIX + router_id
        extids = {ovn_const.OVN_PORT_NAME_EXT_ID_KEY: name}

        for row in ports:
            if self._match_extids(row, extids):
                return row

    def _find_transit_router_port(self, router_id, network_name, ports):
        extids = {
            ovn_const.OVN_DEVID_EXT_ID_KEY: router_id,
            ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY: 'network:router_interface',
            ovn_const.OVN_NETWORK_NAME_EXT_ID_KEY: network_name,
        }
        for row in ports:
            if self._match_extids(row, extids):
                return row

    def _find_vpn_gw_port(self, router_id, ports):
        name = ovn_ipsec.VPN_GW_PORT_PREFIX + router_id
        extids = {ovn_const.OVN_PORT_NAME_EXT_ID_KEY: name}

        for row in ports:
            if self._match_extids(row, extids):
                return row

    def _find_lrouter_by_neutron_id(self, router_id):
        for row in self.nb_api._tables['Logical_Router'].rows.values():
            if row.name == "neutron-" + router_id:
                return row

    def test_agent(self):
        chassis_row = self.sb_api.db_find(
            self.agent_chassis_table,
            ('name', '=', self.chassis_name)).execute(
            check_error=True)[0]

        # Assert that, prior to creating a resource the VPN agent
        # didn't populate the external_ids from the Chassis
        self.assertNotIn(vpn_const.OVN_AGENT_VPN_SB_CFG_KEY,
                         chassis_row['external_ids'])

        # Let's list the agents to force the nb_cfg to be bumped on NB
        # db, which will automatically increment the nb_cfg counter on
        # NB_Global and make ovn-controller copy it over to SB_Global. Upon
        # this event, VPN agent will update the external_ids on its
        # Chassis row to signal that it's healthy.

        row_event = VPNAgentHealthEvent(self.chassis_name, 1,
                                        self.agent_chassis_table)
        self.handler.watch_event(row_event)
        self.new_list_request('agents').get_response(self.api)

        # If we do not time out waiting for the event, then we are assured
        # that the VPN agent has populated the external_ids from the
        # chassis with the nb_cfg, 1 revisions when listing the agents.
        self.assertTrue(row_event.wait())

    def test_service(self):
        r = self.new_list_request('agents').get_response(self.api)
        ext_net, ext_sub = self._make_ext_network()

        server = ovn_ipsec.IPsecVpnOvnDriverCallBack(self.vpn_service_driver)

        # Mock the controller side RPC client (prepare and cast)
        # to be able to check that "vpnservice_updated" will be called
        prepare_mock = mock.Mock()
        prepared_mock = mock.Mock()
        self.vpn_service_driver.agent_rpc.client.prepare = prepare_mock
        prepare_mock.return_value = prepared_mock

        # Create a site (router, network, subnet, vpnservice, site conn)
        site = OvnSiteInfo(self, 1, ext_net, ext_sub)
        site.create_base()
        site.create_vpnservice()
        site.create_site_connection(PEER_ADDR, str(PEER_NET))

        # Check that the vpnservice_updated RPC was triggered towards
        # the agent
        prepare_mock.assert_called_once_with(
            server=self.FAKE_CHASSIS_HOST,
            version=self.vpn_service_driver.agent_rpc.target.version)
        prepared_mock.cast.assert_called_once_with(
            self.context, 'vpnservice_updated',
            router={'id': site.router['id']})

        # Mock the agent->controller RPCs. Let them return data from the
        # actual VPN plugin
        def get_vpn_services_on_host(ctx, host):
            r = server.get_vpn_services_on_host(self.context, host)
            return r

        def get_vpn_transit_network_details(router_id):
            return server.get_vpn_transit_network_details(
                self.context, router_id)

        def get_subnet_info(subnet_id):
            return server.get_subnet_info(self.context, subnet_id)

        r = self.agent_driver.agent_rpc
        r.get_vpn_services_on_host.side_effect = get_vpn_services_on_host
        r.get_vpn_transit_network_details.side_effect = \
            get_vpn_transit_network_details
        r.get_subnet_info.side_effect = get_subnet_info

        # Call the agent's vpnservice_updated as if it was coming from
        # the controller.
        for driver in self.agent.device_drivers:
            driver.vpnservice_updated(driver.context,
                                      router={'id': site.router['id']})

        # Check that transit network and VPN gateway port are set up correctly
        # - transit network exists
        # - router port in transit network exists
        # - transit network port to be bound to chassis exists and
        #   host is assigned
        # - VPN gateway port exists and host is assigned
        # - static route exists towards peer CIDR

        # expect transit network in NB
        transit_row = self._find_transit_lswitch(site.router['id'])
        self.assertIsNotNone(transit_row)

        # check the transit network router port exists
        transit_router_port = self._find_transit_router_port(
            site.router['id'], transit_row.name, transit_row.ports)
        self.assertIsNotNone(transit_router_port)

        # check that the namespace port in the transit network exists
        transit_ns_port = self._find_transit_ns_port(site.router['id'],
                                                     transit_row.ports)
        self.assertIsNotNone(transit_ns_port)

        # check that the port has the requested-host option
        requested_host = transit_ns_port.options.get(
            ovn_const.LSP_OPTIONS_REQUESTED_CHASSIS_KEY)
        self.assertEqual(requested_host, self.FAKE_CHASSIS_HOST)

        # get vpn gateway port via external network lswitch
        ext_row = self._find_lswitch_by_neutron_name("external-net")
        self.assertIsNotNone(ext_row)

        vpn_gw_port = self._find_vpn_gw_port(site.router['id'], ext_row.ports)
        self.assertIsNotNone(vpn_gw_port)
        # check that vpn gateway port has the requested-host option
        requested_host = vpn_gw_port.options.get(
            ovn_const.LSP_OPTIONS_REQUESTED_CHASSIS_KEY)
        self.assertEqual(requested_host, self.FAKE_CHASSIS_HOST)

        # check that static route towards peer cidr is set
        router_row = self._find_lrouter_by_neutron_id(site.router['id'])
        self.assertIsNotNone(router_row)
        for r in router_row.static_routes:
            if r.ip_prefix == str(PEER_NET):
                route = r
                break
        else:
            route = None

        self.assertIsNotNone(route)
        self.assertEqual(route.nexthop, ovn_ipsec.VPN_TRANSIT_RIP)

        # Check agent side
        # - network namespace
        # - routes towards transit network's gateway IP
        # - devices and their IP addresses in the namespace
        ns_name = self.ns_prefix + site.router['id']
        devlen = lib_constants.LINUX_DEV_LEN
        transit_dev = ('vr' + transit_ns_port.name)[:devlen]
        gw_dev = ('vg' + vpn_gw_port.name)[:devlen]
        self.assertTrue(ip_lib.network_namespace_exists(ns_name))
        device = ip_lib.IPDevice(None, namespace=ns_name)
        routes = device.route.list_routes(lib_constants.IP_VERSION_4,
                                          proto='static',
                                          via=ovn_ipsec.VPN_TRANSIT_LIP)
        self.assertEqual(len(routes), 1)
        self.assertEqual(routes[0]['via'], ovn_ipsec.VPN_TRANSIT_LIP)
        self.assertEqual(routes[0]['cidr'], site.local_cidr)
        self.assertEqual(routes[0]['device'], transit_dev)

        # check addresses in namespace
        addrs = device.addr.list(ip_version=lib_constants.IP_VERSION_4)
        addrs_dict = {a['name']: a for a in addrs}
        self.assertIn(transit_dev, addrs_dict)
        self.assertEqual(
            addrs_dict[transit_dev]['cidr'],
            transit_ns_port.external_ids[ovn_const.OVN_CIDRS_EXT_ID_KEY])

        self.assertIn(gw_dev, addrs_dict)
        self.assertEqual(
            addrs_dict[gw_dev]['cidr'],
            vpn_gw_port.external_ids[ovn_const.OVN_CIDRS_EXT_ID_KEY])
